@@ -275,6 +275,7 @@ static char *keyType(CMPIType type)
 
 CMPIType guessType(char *val)
 {
+   if (!val) return CMPI_null;
    if (((*val=='-' || *val=='+') && strlen(val)>1) || isdigit(*val)) {
       char *c;
       for (c=val+1; ; c++) {
@@ -294,6 +295,7 @@ CMPIValue makeFromEmbeddedObject(XtokValue value, char *ns)
 {
 	XtokProperty *p = NULL;
 	CMPIValue val;
+	memset(&val, 0, sizeof(CMPIValue));
 	CMPIObjectPath *path;
 	
 	if(value.type == typeValue_Instance) {
@@ -345,7 +347,7 @@ CMPIValue str2CMPIValue(CMPIType type, XtokValue val, XtokValueReference *ref, c
      value.array = TrackedCMPIArray(max,t,NULL);
      if (value.array != NULL) {
        for (i=0; i<max; i++) {
-	 v = str2CMPIValue(t, arr->values[i], refarr->values+i,NULL);
+	 v = str2CMPIValue(t, arr->values[i], refarr->values+i,ns);
 	 CMSetArrayElementAt(value.array, i, &v, t); 
        }
        return value;
@@ -551,7 +553,12 @@ static int nsPath2xml(CMPIObjectPath * ci, UtilStringBuffer * sb)
    if (hn && *hn) {
       sb->ft->appendChars(sb, hn);
    } else {
-      SFCB_APPENDCHARS_BLOCK(sb, "localhost");
+      hn = malloc(HOST_NAME_MAX);
+      if(gethostname(hn, HOST_NAME_MAX) == 0) {
+        sb->ft->appendChars(sb, hn);
+      } else {
+        SFCB_APPENDCHARS_BLOCK(sb, "localhost");
+      }
    }
    SFCB_APPENDCHARS_BLOCK(sb, "</HOST>\n");
    
@@ -629,18 +636,25 @@ void data2xml(CMPIData * data, void *obj, CMPIString * name, CMPIString * refNam
       sb->ft->appendBlock(sb, bTag,bTagLen);
       sb->ft->appendChars(sb, (char *) name->hdl);
       if (param) SFCB_APPENDCHARS_BLOCK(sb, "\" PARAMTYPE=\"");
-      else if (bTag) {
-        SFCB_APPENDCHARS_BLOCK(sb, "\" TYPE=\"");
-        if(data->type & CMPI_instance || data->type & CMPI_class) {
-             SFCB_APPENDCHARS_BLOCK(sb, "string");
-        } else {
-         sb->ft->appendChars(sb, dataType(data->type));
-        }
+      else if (bTag) SFCB_APPENDCHARS_BLOCK(sb, "\" TYPE=\"");
+      if (data->type == CMPI_refA) {
+         SFCB_APPENDCHARS_BLOCK(sb, "reference");
+      }
+      else if(((data->type & ~CMPI_ARRAY) == CMPI_instance) 
+               || ((data->type & ~CMPI_ARRAY) == CMPI_class)) {
+         SFCB_APPENDCHARS_BLOCK(sb, "string\" EmbeddedObject=\"object");
+      } else {
+          sb->ft->appendChars(sb, dataType(data->type));
       }
       SFCB_APPENDCHARS_BLOCK(sb, "\">\n");
       if (qsb) sb->ft->appendChars(sb, (char *) qsb->hdl);
       if (data->state == 0) {
-         SFCB_APPENDCHARS_BLOCK(sb, "<VALUE.ARRAY>\n");
+         if (data->type == CMPI_refA) {
+            SFCB_APPENDCHARS_BLOCK(sb, "<VALUE.REFARRAY>\n");
+         }
+         else {
+            SFCB_APPENDCHARS_BLOCK(sb, "<VALUE.ARRAY>\n");
+         }
          for (j = 0; j < ac; j++) {
             d = CMGetArrayElementAt(ar, j, NULL);
 	    if ((d.state & CMPI_nullValue)==0) {
@@ -651,7 +665,11 @@ void data2xml(CMPIData * data, void *obj, CMPIString * name, CMPIString * refNam
 	      }
 	    }
          }
-         SFCB_APPENDCHARS_BLOCK(sb, "</VALUE.ARRAY>\n");
+         if (data->type == CMPI_refA) {
+            SFCB_APPENDCHARS_BLOCK(sb, "</VALUE.REFARRAY>\n");
+         } else {
+            SFCB_APPENDCHARS_BLOCK(sb, "</VALUE.ARRAY>\n");
+         }
       }
    }
    
@@ -674,7 +692,7 @@ void data2xml(CMPIData * data, void *obj, CMPIString * name, CMPIString * refNam
       }
       
       else if (*type == '%') {                  
-	 sb->ft->appendBlock(sb, bTag, bTagLen);
+	     sb->ft->appendBlock(sb, bTag, bTagLen);
          sb->ft->appendChars(sb, (char *) name->hdl);
          SFCB_APPENDCHARS_BLOCK(sb, "\" EmbeddedObject=\"object");
          if (param) SFCB_APPENDCHARS_BLOCK(sb, "\" PARAMTYPE=\"string\">\n");
@@ -689,14 +707,12 @@ void data2xml(CMPIData * data, void *obj, CMPIString * name, CMPIString * refNam
      }
       
       else {
-	 sb->ft->appendBlock(sb, bTag, bTagLen);
+	     sb->ft->appendBlock(sb, bTag, bTagLen);
          sb->ft->appendChars(sb, (char *) name->hdl);
          if (param) SFCB_APPENDCHARS_BLOCK(sb, "\" PARAMTYPE=\"");
-         else  if (bTag) {
-            SFCB_APPENDCHARS_BLOCK(sb, "\" TYPE=\"");
-            sb->ft->appendChars(sb, type);
-            SFCB_APPENDCHARS_BLOCK(sb, "\">\n");
-         }
+         else  if (bTag) SFCB_APPENDCHARS_BLOCK(sb, "\" TYPE=\"");
+         sb->ft->appendChars(sb, type);
+         SFCB_APPENDCHARS_BLOCK(sb, "\">\n");
          if (qsb) sb->ft->appendChars(sb, (char *) qsb->hdl);
          if (data->state == 0) value2xml(*data, sb, 1);
       }
@@ -1009,8 +1025,6 @@ int enum2xml(CMPIEnumeration * enm, UtilStringBuffer * sb, CMPIType type,
    CMPIObjectPath *cop;
    CMPIInstance *ci;
    CMPIConstClass *cl;
-   CMPIString *hs;
-   static char hss[HOST_NAME_MAX+1];
 
    _SFCB_ENTER(TRACE_CIMXMLPROC, "enum2xml");
    
@@ -1035,15 +1049,6 @@ int enum2xml(CMPIEnumeration * enm, UtilStringBuffer * sb, CMPIType type,
       else if (type == CMPI_instance) {
          ci = CMGetNext(enm, NULL).value.inst;
          cop = CMGetObjectPath(ci, NULL);
-	 hs = CMGetHostname(cop,NULL);
-	 if (hs == NULL || CMGetCharPtr(hs)==NULL) {
-	   /* required for value.objectwithpath */
-	   if(!hss[0] && gethostname(hss, HOST_NAME_MAX)) {
-          CMSetHostname(cop,"localhost");
-       } else {
-          CMSetHostname(cop,hss);
-       }
-	 }
          if (xmlAs==XML_asObj) {
             SFCB_APPENDCHARS_BLOCK(sb, "<VALUE.OBJECTWITHPATH>\n");
             SFCB_APPENDCHARS_BLOCK(sb, "<INSTANCEPATH>\n");
