@@ -1,6 +1,6 @@
 
 /*
- * $Id: cimXmlParser.c,v 1.25 2007/12/21 16:04:48 sschuetz Exp $
+ * $Id: cimXmlParser.c,v 1.33 2009/06/05 20:09:34 buccella Exp $
  *
  *  © Copyright IBM Corp. 2005, 2007
  *
@@ -42,7 +42,7 @@ typedef struct tags {
 } Tags;
 
 
-
+/* TODO: be more graceful than just exit()  */
 static void Throw(XmlBuffer * xb, char *msg)
 {
    printf("*** Error: %s\n", msg);
@@ -86,6 +86,10 @@ static int getChar(XmlBuffer * xb, const char c)
    return 0;
 }
 
+/* skip nulled char and whitespace, 
+   see if we're on a new tag and return a pointer, 
+   or NULL otherwise
+ */
 static char *nextTag(XmlBuffer * xb)
 {
    if (xb->nulledChar) {
@@ -98,6 +102,10 @@ static char *nextTag(XmlBuffer * xb)
    return NULL;
 }
 
+/* compares two strings
+   return 1: n begins with t and following char is alphanumeric
+   return 0: otherwise
+ */
 static int nextEquals(const char *n, const char *t)
 {
    int l = strlen(t);
@@ -109,6 +117,7 @@ static int nextEquals(const char *n, const char *t)
    return 0;
 }
 
+/* scan until we find the end of the tag ('>') */
 static char skipTag(XmlBuffer * xb)
 {
    while (*xb->cur != '>' && xb->last > xb->cur)
@@ -218,18 +227,23 @@ static int attrsOk(XmlBuffer * xb, const XmlElement * e, XmlAttr * r,
                    const char *tag, int etag)
 {
    unsigned int n;
-   char *ptr, wa[32];
+#define MAXATTR 32
+#define WORDLEN 20
+   char *ptr, wa[MAXATTR];
    char msg1[] = { "Unknown attribute in list for " };
    char msg2[] = { "Bad attribute list for " };
-   char word[11];
+   char word[WORDLEN+1];
 
-   for (n = 0; (e + n)->attr; n++)
-      wa[n] = 0;
+   for (n = 0; (e + n)->attr; n++) {
+      if (n >= MAXATTR)
+        Throw(xb, "Too many attribute in XML");
+        wa[n] = 0;
+   }
 
    xb->eTagFound = 0;
    for (skipWS(xb); isalpha(*xb->cur); skipWS(xb)) {
 //      for (n=0; n < a.size(); n++) {
-      for (n = 0; (e + n)->attr; n++) {
+      for (n = 0; (e + n)->attr; n++) {  /* MAXATTR checked above */
          if (wa[n] == 1)
             continue;
          if (getWord(xb, (e + n)->attr, 0)) {
@@ -245,14 +259,14 @@ static int attrsOk(XmlBuffer * xb, const XmlElement * e, XmlAttr * r,
             }
          }
       }
-      strncpy(word, xb->cur, 10);
-      word[10] = 0;
-      ptr = (char *) alloca(strlen(tag) + strlen(msg1) + 8 + 20);
+      strncpy(word, xb->cur, WORDLEN);
+      word[WORDLEN] = 0;
+      ptr = (char *) alloca(strlen(tag) + strlen(msg1) + WORDLEN + sizeof(char)*4);
       strcpy(ptr, msg1);
       strcat(ptr, tag);
-      strcat(ptr, " (");
+      strcat(ptr, " ("); /* 2 chars */
       strcat(ptr, word);
-      strcat(ptr, ")");
+      strcat(ptr, ")"); /* 1 char + \0 */
       Throw(xb, ptr);
     ok:;
    }
@@ -272,15 +286,19 @@ static int attrsOk(XmlBuffer * xb, const XmlElement * e, XmlAttr * r,
       return 1;
    }
 
-   ptr = (char *) alloca(strlen(tag) + strlen(msg2) + 48);
+   /* build error message for Throw(): "Bad attribute list for: <TAG>: <chars...> */
+   ptr = (char*)alloca(strlen(msg2) + strlen(tag) + sizeof(char)*3 + strlen(word));
    strcpy(ptr, msg2);
    strcat(ptr, tag);
    strcat(ptr, ": ");
-   strncpy(word, xb->cur, 10);
-   strcat(ptr, word);
-   strcat(ptr, tag);
+   /* ensure we have at least WORDLEN chars left in the XML */
+   int wlen = (xb->cur < (xb->last - WORDLEN)) ? WORDLEN : (xb->last - xb->cur); 
+   strncpy(word, xb->cur, wlen);
+   strncat(ptr, word, wlen);
    Throw(xb, ptr);
    return -1;
+#undef WORDLEN
+#undef MAXATTR
 }
 
 /* Is this Broken?  I guess we don't allow escaping the quotes */
@@ -312,34 +330,117 @@ static struct _xmlescape {
     { "&amp;", '&', 5 },
     { "&gt;", '>', 4 },
     { "&lt;", '<', 4 },
-    { "&#x20;", ' ', 6 },
 };
 
+// arbitrary - esc chars are only valid in format &#<dec> or &#x<hex>
+#define MAX_UNESC_CHAR_LEN 10
+
 static int xmlUnescape(char *buf, char *end)
 {
-  int i;
-  for (i=0; i < sizeof(xmlEscapeTab)/sizeof(struct _xmlescape);i++) {
-    if (end-buf >= xmlEscapeTab[i].len && 
-	strncmp(buf,xmlEscapeTab[i].escaped,xmlEscapeTab[i].len) == 0) {
-      *buf = xmlEscapeTab[i].unescaped;
-      memmove(buf+1,buf+xmlEscapeTab[i].len,end-buf-xmlEscapeTab[i].len+1);
-      return xmlEscapeTab[i].len-1;
+    char escchar[MAX_UNESC_CHAR_LEN] = {0};
+    char *semiloc = NULL;
+    char *amploc = NULL;
+
+    semiloc = strchr((char *)(buf+2), ';');
+    if (semiloc==NULL)
+    {
+        // either malformed escape sequence, or non-escape sequence
+        // we didn't do anything, return 0
+        return 0;
     }
-  }
-  return 0;
+    // now see if I have another & before the first semicolon-
+    // if I do, either this is not an escape sequence, or it's mal-formed
+    amploc = strchr((char *)(buf+2), '&');
+    if (amploc && (amploc < semiloc))
+    {
+        // either malformed escape sequence, or non-escape sequence
+        // we didn't do anything, return 0
+        return 0;
+    }
+    if (semiloc-buf > MAX_UNESC_CHAR_LEN) 
+    {
+        // if there's more than 10 characters in the unescaped string, then
+        // it is either not an escaped char or it is malformed...  just return
+        return 0;
+    }
+
+    if (buf[1] == '#')
+    {
+        // got a generic escaped character:  should follow format:
+        // &#x<hex> or &#<dec>
+        // copy from original string, skipping '&#'
+        strncpy(escchar, buf+2, (int)(semiloc-buf-2));
+
+        int unescchar = 0;
+        int charlen = 0;
+        int diflen = 0; // total unescaped char len, with &#<x><val>;
+        if (escchar[0]=='x' || escchar[0]=='X')
+        {
+            // expecting a hex number, convert accordingly
+            charlen=strlen(escchar)-1;//x
+            diflen = charlen+4; //&#x and ;
+            if (0 == sscanf(&escchar[1], "%x", &unescchar))
+            {
+                // couldn't convert hex number - malformed
+                return 0;
+            }
+        }
+        else
+        {
+            // expecting a decimal number, convert accordingly
+            charlen=strlen(escchar);
+            diflen = charlen+3; //&# and ;
+            if (0 == sscanf(escchar, "%d", &unescchar))
+            {
+                // couldn't convert decimal number - malformed
+                return 0;
+            }
+        }
+        // move the unescaped character to first position
+        *buf = unescchar;
+        // now move rest of string, after the escaped character we just converted, to follow
+        // newly placed unescaped character
+        memmove(buf+1, buf+diflen,(int)(end-buf-diflen+1));
+        // and return the number of characters we removed from the total string len
+        return diflen-1;
+    }
+    else
+    {
+        int i=0;
+        // expecting a known/defined escape character, ie &lt;
+        strncpy(escchar, buf, (int)(semiloc-buf+1));
+        for (i=0; i < sizeof(xmlEscapeTab)/sizeof(struct _xmlescape);i++) 
+        {
+            if (strncmp(escchar, xmlEscapeTab[i].escaped, xmlEscapeTab[i].len) == 0)
+            {
+                *buf = xmlEscapeTab[i].unescaped;
+                memmove(buf+1,buf+xmlEscapeTab[i].len,end-buf-xmlEscapeTab[i].len+1);
+                //return the number of characters we removed
+                return xmlEscapeTab[i].len-1;
+            }
+        }
+    }
+    return 0;
 }
 
 static char *getContent(XmlBuffer * xb)
 {
    char *start = xb->cur,*end, *help;
+
+    /* if this is an ending tag, return NULL (no content) */
     if (xb->eTagFound)
       return NULL;
+
+   /* scan through until we hit an '<' or end of buffer*/
    for (; *xb->cur != '<' && xb->cur < xb->last; xb->cur++);
+
+   /* store the char we found, set it to null to use as a marker in the code below */
    xb->nulledChar = *(xb->cur);
    *(xb->cur) = 0;
+
    /* skip leading blanks */
    while (*start && *start<=' ') start++;
-   if (start == NULL) return "";
+   if (start == NULL) return "";  // should check for *start == NULL instead? start==NULL shouldn't happen, ever.
    end=xb->cur;
    /* strip trailing blanks */
    while (*(end-1)<=' ') {
@@ -616,6 +717,7 @@ static int procParamValue(YYSTYPE * lvalp, ParserControl * parm)
                   break;
                }
             }
+	    // is this right?  why isn't reference covered by the types array in the loop above?
             if (lvalp->xtokParamValue.type==0) {
                if (strcasecmp(attr[1].attr, "reference") == 0) 
                   lvalp->xtokParamValue.type = CMPI_ref;
@@ -625,6 +727,19 @@ static int procParamValue(YYSTYPE * lvalp, ParserControl * parm)
             if(strcasecmp(attr[2].attr, "instance") == 0
                || strcasecmp(attr[2].attr, "object") == 0) {
                lvalp->xtokParamValue.type = CMPI_instance;
+
+               /* unescape */
+               char *help = parm->xmb->cur;
+               char *end = strstr(help, "</PARAMVALUE");
+               while (help < end) {
+                 if (*help == '&') {
+                   int fewer = xmlUnescape(help, end);
+                   end -= fewer;
+                   memset(end, ' ', fewer);
+                 }
+                 help += 1;
+               }
+
             } else {
                Throw(NULL, "Invalid value for attribute EmbeddedObject");
             }
@@ -826,7 +941,7 @@ static int procValue(YYSTYPE * lvalp, ParserControl * parm)
    if (tagEquals(parm->xmb, "VALUE")) {
       char *v;
       if (attrsOk(parm->xmb, elm, attr, "VALUE", ZTOK_VALUE)) {
-         v=getContent(parm->xmb);
+         v=getContent(parm->xmb); /* v is a pointer to where the content starts*/
          lvalp->xtokValue.value = v;
          return XTOK_VALUE;
       }
@@ -1431,11 +1546,14 @@ int sfcXmllex(YYSTYPE * lvalp, ParserControl * parm)
          _SFCB_RETURN(0);
       }   
       _SFCB_TRACE(1, ("--- token: %.32s\n",next));
+
+      /* found a tag, reset eTagFound flag, return the token for the tag that was found */
       if (parm->xmb->eTagFound) {
          parm->xmb->eTagFound = 0;
          _SFCB_RETURN(parm->xmb->etag);
       }
 
+      /* found an ending tag, loop from 0-#ofTags, return the token for the matching opening tag */
       if (*next == '/') {
 	for (i = 0, m = sizeof(tags)/sizeof(Tags); i < m; i++) {
             if (nextEquals(next + 1, tags[i].tag) == 1) {
@@ -1446,14 +1564,16 @@ int sfcXmllex(YYSTYPE * lvalp, ParserControl * parm)
       }
 
       else {
+	/* skip comment section */
          if (strncmp(parm->xmb->cur, "<!--", 4) == 0) {
             parm->xmb->cur = strstr(parm->xmb->cur, "-->") + 3;
             continue;
          }
          for (i = 0, m = sizeof(tags)/sizeof(Tags); i < m; i++) {
             if (nextEquals(next, tags[i].tag) == 1) {
-   //printf("+++ %d\n",i);
-               rc=tags[i].process(lvalp, parm);
+
+               // fprintf(stderr, " sfcXmllex| processing for: %.10s\n", next);
+               rc=tags[i].process(lvalp, parm); /* call a procXXX fn based on tag name */
                _SFCB_RETURN(rc);
             }
          }
@@ -1708,9 +1828,18 @@ static void freeInstance(XtokInstance *op)
 
 static  void freeNewValue(XtokNewValue *op)
 {
+  switch(op->type) {
+  case typeValue_Instance:
+/* case typeValue_Class: */
     freeValue(&op->val);
+    break;
+  case CMPI_ref:
     freeReference(&op->ref);
+    break;
+  case CMPI_ARRAY:
     freeArray(&op->arr);
+    break;
+  }
 }
 
 static void freeMethodCall(XtokMethodCall* op)
