@@ -1,6 +1,6 @@
 
 /*
- * $Id: cimXmlRequest.c,v 1.48 2008/01/30 22:20:26 buccella Exp $
+ * $Id: cimXmlRequest.c,v 1.55 2009/06/05 19:18:24 buccella Exp $
  *
  * © Copyright IBM Corp. 2005, 2007
  *
@@ -44,6 +44,7 @@
 
 #include "queryOperation.h"
 #include "config.h"
+#include "control.h"
 
 #ifdef SFCB_IX86
 #define SFCB_ASM(x) asm(x)
@@ -51,6 +52,11 @@
 #define SFCB_ASM(x)
 #endif
 
+#ifdef LOCAL_CONNECT_ONLY_ENABLE
+// from httpAdapter.c
+int noChunking = 0;
+#endif // LOCAL_CONNECT_ONLY_ENABLE
+
 typedef struct handler {
    RespSegments(*handler) (CimXmlRequestContext *, RequestHdr * hdr);
 } Handler;
@@ -266,25 +272,26 @@ UtilStringBuffer *segments2stringBuffer(RespSegment *rs)
    return sb;
 }
 
+
 static char *getErrSegment(int rc, char *m)
 {
-   char msg[1024];
+   char *msg;
    char *escapedMsg;
    
    if (m && *m) {
      escapedMsg = XMLEscape(m, NULL);
-       snprintf(msg, sizeof(msg), "<ERROR CODE=\"%d\" DESCRIPTION=\"%s\"/>\n",
+       msg = sfcb_snprintf("<ERROR CODE=\"%d\" DESCRIPTION=\"%s\"/>\n",
                 rc, escapedMsg);
        free(escapedMsg);
    }
    else if (rc > 0 && rc < 18) {
-       snprintf(msg, sizeof(msg), "<ERROR CODE=\"%d\" DESCRIPTION=\"%s\"/>\n",
+       msg = sfcb_snprintf("<ERROR CODE=\"%d\" DESCRIPTION=\"%s\"/>\n",
                 rc, cimMsg[rc]);
    }
    else  {
-       snprintf(msg, sizeof(msg), "<ERROR CODE=\"%d\"/>\n", rc);
+       msg = sfcb_snprintf("<ERROR CODE=\"%d\"/>\n", rc);
    }
-   return strdup(msg);
+   return msg; 
 }
 /*
 static char *getErrorSegment(CMPIStatus rc)
@@ -297,13 +304,13 @@ static char *getErrorSegment(CMPIStatus rc)
 */
 char *getErrTrailer(int id, int rc, char *m)
 {
-   char msg[1024];
+   char *msg;
    
-   if (m && *m) snprintf(msg, sizeof(msg), "CIMStatusCodeDescription: %s\r\n",m);
+   if (m && *m) msg = sfcb_snprintf("CIMStatusCodeDescription: %s\r\n",m);
    else if (rc > 0 && rc < 18)
-      snprintf(msg, sizeof(msg), "CIMStatusCodeDescription: %s\r\n",cimMsg[rc]);
-   else snprintf(msg, sizeof(msg), "CIMStatusCodeDescription: *Unknown*\r\n");
-   return strdup(msg);
+      msg = sfcb_snprintf("CIMStatusCodeDescription: %s\r\n",cimMsg[rc]);
+   else msg = strdup("CIMStatusCodeDescription: *Unknown*\r\n");
+   return msg; 
 }
 
 
@@ -347,31 +354,38 @@ static RespSegments ctxErrResponse(RequestHdr * hdr,
 {
    MsgXctl *xd = ctx->ctlXdata;
    char msg[256];
+   CMPIrc err;
 
    switch (ctx->rc) {
    case MSG_X_NOT_SUPPORTED:
       hdr->errMsg = strdup("Operation not supported yy");
+      err = CMPI_RC_ERR_NOT_SUPPORTED;
       break;
    case MSG_X_INVALID_CLASS:
       hdr->errMsg = strdup("Class not found");
+      err = CMPI_RC_ERR_INVALID_CLASS;
       break;
    case MSG_X_INVALID_NAMESPACE:
       hdr->errMsg = strdup("Invalid namespace");
+      err = CMPI_RC_ERR_INVALID_NAMESPACE;
       break;
    case MSG_X_PROVIDER_NOT_FOUND:
       hdr->errMsg = strdup("Provider not found or not loadable");
+      err = CMPI_RC_ERR_NOT_FOUND;
       break;
    case MSG_X_FAILED:
       hdr->errMsg = strdup(xd->data);
+      err = CMPI_RC_ERR_FAILED; 
       break;
    default:
       sprintf(msg, "Internal error - %d\n", ctx->rc);
       hdr->errMsg = strdup(msg);
+      err = CMPI_RC_ERR_FAILED;
    }
    if (meth) return
-      methodErrResponse(hdr,getErrSegment(CMPI_RC_ERR_INVALID_CLASS,hdr->errMsg));
+      methodErrResponse(hdr,getErrSegment(err,hdr->errMsg));
    return
-      iMethodErrResponse(hdr,getErrSegment(CMPI_RC_ERR_INVALID_CLASS,hdr->errMsg));
+      iMethodErrResponse(hdr,getErrSegment(err,hdr->errMsg));
 };
 
 
@@ -744,6 +758,7 @@ static RespSegments createClass(CimXmlRequestContext * ctx, RequestHdr * hdr)
    CMPIObjectPath *path;
    CMPIConstClass cls;
    ClClass *cl;
+   ClClass *tmp;
    int irc;
    BinRequestContext binCtx;
    BinResponseHdr *resp;
@@ -857,7 +872,9 @@ static RespSegments createClass(CimXmlRequestContext * ctx, RequestHdr * hdr)
       }
    }
    
+   tmp = cl;
    cl = ClClassRebuildClass(cl,NULL); 
+   free(tmp);
    cls=initConstClass(cl);
 
    sreq.principal = setCharsMsgSegment(ctx->principal);
@@ -1938,6 +1955,89 @@ static RespSegments references(CimXmlRequestContext * ctx, RequestHdr * hdr)
    _SFCB_RETURN(ctxErrResponse(hdr, &binCtx,0));
 }
 
+int updateMethodParamTypes(RequestHdr *hdr) {
+
+  _SFCB_ENTER(TRACE_CIMXMLPROC, "updateMethodParamTypes");
+
+  CMPIConstClass *cls = NULL;
+  ClMethod *meth;
+  ClParameter *param = NULL;
+  int i, m;
+  ClClass *cl;
+  char *mname;
+  XtokParamValue *ptok;
+  int p, pm;
+
+  XtokMethodCall *req = (XtokMethodCall *) hdr->cimRequest;
+  cls = getConstClass((char*)req->op.nameSpace.data, (char*)req->op.className.data);
+  if (!cls) {
+    _SFCB_RETURN(CMPI_RC_ERR_INVALID_CLASS);
+  }
+
+  cl = (ClClass *) cls->hdl;
+
+  /* check that the method specified in req exists in class */ 
+  for (i = 0, m = ClClassGetMethodCount(cl); i < m; i++) {
+    ClClassGetMethodAt(cl, i, NULL, &mname, NULL);
+    if (strcasecmp(req->method, mname) == 0) {
+      break;
+    }
+  }
+  if (i == m) {
+    _SFCB_RETURN(CMPI_RC_ERR_METHOD_NOT_FOUND);
+  }
+
+  meth=((ClMethod*)ClObjectGetClSection(&cl->hdr,&cl->methods))+i;
+
+  /* loop through all params from parsed req */
+  for (ptok = req->paramValues.first; ptok; ptok = ptok->next) {
+    CMPIParameter pdata; 
+    char *sname;
+
+    /* loop through all params for meth */
+    for (p = 0, pm = ClClassGetMethParameterCount(cl, i); p < pm; p++) {
+      ClClassGetMethParameterAt(cl, meth, p, &pdata, &sname);
+
+      if (strcasecmp(sname, ptok->name) == 0) {
+        // fprintf(stderr, "%s matches %s", sname, ptok->name);
+        param=((ClParameter*)ClObjectGetClSection(&cl->hdr,&meth->parameters))+p;
+        break;
+      }
+    }
+    if (p == pm) {
+      _SFCB_RETURN(CMPI_RC_ERR_INVALID_PARAMETER);
+    }
+
+    // fprintf(stderr, " pdata.type=%u (expec), ptok->type=%u\n", pdata.type, ptok->type);
+    /* special case: EmbeddedInstance. Parser will set type to instance, but repository
+       would have type as string. Check here to not fail the else-if below. */
+    if (param && (ptok->type & CMPI_instance)) {
+      int isEI = 0;
+      int qcount = ClClassGetMethParmQualifierCount(cl, meth, i);
+      for (; qcount > 0; qcount--) {
+        char* qname;
+        ClClassGetMethParamQualifierAt(cl, param, qcount, NULL, &qname);
+        if (strcmp(qname, "EmbeddedInstance") == 0) {
+	  // fprintf(stderr, "  is EmbeddedInstance\n");
+          isEI = 1;
+          break;
+        }
+      }
+      if (isEI) continue;
+    }
+
+    if (ptok->type==0) {
+      /* Type was unknown, fill it in */
+      // printf("parameter %s missing type, using %s\n", sname, paramType(pdata.type));
+      ptok->type = pdata.type;
+    } else if (ptok->type != pdata.type) {
+      /* Parameter type mismatch */
+      _SFCB_RETURN(CMPI_RC_ERR_TYPE_MISMATCH);
+    } 
+  }
+  _SFCB_RETURN(CMPI_RC_OK);
+}
+ 
 static RespSegments invokeMethod(CimXmlRequestContext * ctx, RequestHdr * hdr)
 {
    _SFCB_ENTER(TRACE_CIMXMLPROC, "invokeMethod");
@@ -1946,7 +2046,7 @@ static RespSegments invokeMethod(CimXmlRequestContext * ctx, RequestHdr * hdr)
    CMPIType type;
    CMPIValue val, *valp;
    UtilStringBuffer *sb;
-   int irc, i, m;
+   int irc, i, m, rc, vmpt = 0;
    BinRequestContext binCtx;
    BinResponseHdr *resp;
    RespSegments rsegs;
@@ -1970,14 +2070,26 @@ static RespSegments invokeMethod(CimXmlRequestContext * ctx, RequestHdr * hdr)
    sreq.principal = setCharsMsgSegment(ctx->principal);
    sreq.hdr.sessionId=ctx->sessionId;
 
+   if (getControlBool("validateMethodParamTypes", &vmpt))
+     vmpt=1;
+
    for (p = req->paramValues.first; p; p = p->next) {
-      // this is a problem: - paramvalue without type
-      if (p->type==0) p->type=CMPI_string;
+     /* Update untyped params (p->type==0) and verify those that were specified */
+     if (p->type==0 || vmpt) {
+       rc = updateMethodParamTypes(hdr);
+
+       if (rc != CMPI_RC_OK) {
+	 rsegs = methodErrResponse(hdr, getErrSegment(rc, NULL));
+	 _SFCB_RETURN(rsegs);
+       }
+     }
+      
       if (p->value.value) {
 	CMPIValue val = str2CMPIValue(p->type, p->value, &p->valueRef, req->op.nameSpace.data);
 	CMAddArg(in, p->name, &val, p->type);
       }
-   }   
+   }
+
    sreq.in = setArgsMsgSegment(in);
    sreq.out = setArgsMsgSegment(NULL);
    sreq.method = setCharsMsgSegment(req->method);
@@ -2132,9 +2244,14 @@ static RespSegments setProperty(CimXmlRequestContext * ctx, RequestHdr * hdr)
    else {
       t = req->newVal.type;
    }
-   val = str2CMPIValue(t, req->newVal.val, &req->newVal.ref, req->op.nameSpace.data);
-   
-   CMSetProperty(inst, req->propertyName, &val, t);
+   if (t != CMPI_null) {
+      val = str2CMPIValue(t, req->newVal.val, &req->newVal.ref, req->op.nameSpace.data);
+      CMSetProperty(inst, req->propertyName, &val, t);
+   }
+   else {
+      val.string = 0;
+      CMSetProperty(inst, req->propertyName, 0, t);
+   }
 
    sreq.principal = setCharsMsgSegment(ctx->principal);
    sreq.path = setObjectPathMsgSegment(path);
