@@ -1,6 +1,6 @@
 
 /*
- * $Id: cimcClientSfcbLocal.c,v 1.20 2008/04/16 17:41:29 gdread Exp $
+ * $Id: cimcClientSfcbLocal.c,v 1.29 2009/03/31 21:29:46 buccella Exp $
  *
  * © Copyright IBM Corp. 2006, 2007
  *
@@ -29,7 +29,7 @@
 
  
 #define CIMCSetStatusWithChars(st,rcp,chars) \
-      { if (st) { (st)->rc=(rcp); \
+      { if (st != NULL) { (st)->rc=(rcp); \
         (st)->msg=NewCMPIString((chars),NULL); }}
         
 #define NewCMPIString sfcb_native_new_CMPIString
@@ -45,7 +45,6 @@ extern CMPIObjectPath *relocateSerializedObjectPath(void *area);
 extern CMPIInstance *relocateSerializedInstance(void *area);
 extern CMPIArgs *relocateSerializedArgs(void *area);
 
-extern CMPIString *NewCMPIString(const char*, CMPIStatus*);
 extern CMPIObjectPath *NewCMPIObjectPath(const char*ns, const char* cn, CMPIStatus*);
 extern CMPIInstance *NewCMPIInstance(CMPIObjectPath * cop, CMPIStatus * rc);
 extern CMPIArray *NewCMPIArray(CMPICount size, CMPIType type, CMPIStatus * rc);
@@ -111,7 +110,6 @@ typedef struct thrdInfo {
 	CMPIStatus * rc ;
 } thrdInfo ;
       
-void sfcb_native_array_increase_size(const CMPIArray *, CMPICount);
 static CMPIEnumeration * enumInstances(Client *,CMPIObjectPath *,	CMPIFlags,char **,CMPIStatus *);
                   void * enumInstancesThrd( thrdInfo * ) ;
 static CMPIEnumeration * enumInstanceNames(Client *,CMPIObjectPath *,CMPIStatus *);
@@ -220,40 +218,43 @@ void freeResps(BinResponseHdr **resp, int count)
 }
 
 extern void setEnumArray(CMPIEnumeration * enumeration,CMPIArray * array);
-
 static CMPIEnumeration *cpyEnumResponses(BinRequestContext * binCtx,
                                  BinResponseHdr ** resp,
                                  int arrLen)
 {
    int i, c, j;
-   CMPIInstance *inst;
+   union o {
+     CMPIInstance *inst;
+     CMPIObjectPath *path;
+     CMPIConstClass *cls;
+   } object;
    CMPIArray *ar,*art;
    CMPIEnumeration *enm;
    CMPIStatus rc;
-
+   
    _SFCB_ENTER(TRACE_CIMXMLPROC, "genEnumResponses");
-
+   
    ar = NewCMPIArray(arrLen, binCtx->type, NULL);
    art = NewCMPIArray(0, binCtx->type, NULL);
-
+   
    for (c = 0, i = 0; i < binCtx->rCount; i++) {
       for (j = 0; j < resp[i]->count; c++, j++) {
          if (binCtx->type == CMPI_ref)
-            relocateSerializedObjectPath(resp[i]->object[j].data);
+            object.path = relocateSerializedObjectPath(resp[i]->object[j].data);
          else if (binCtx->type == CMPI_instance)
-            relocateSerializedInstance(resp[i]->object[j].data);
+            object.inst = relocateSerializedInstance(resp[i]->object[j].data);
          else if (binCtx->type == CMPI_class) {
-            relocateSerializedConstClass(resp[i]->object[j].data);
+            object.cls = relocateSerializedConstClass(resp[i]->object[j].data);
          }
-    //     inst=CMClone(inst,NULL);
-         rc=CMSetArrayElementAt(ar,c, (CMPIValue*)&inst, binCtx->type);
+    //     object.inst=CMClone(object.inst,NULL);
+         rc=CMSetArrayElementAt(ar,c, (CMPIValue*)&object.inst, binCtx->type);
       }
    }
 
    enm = NewCMPIEnumeration(art, NULL);
    setEnumArray(enm,ar);
    art->ft->release(art);
-
+   
    _SFCB_RETURN(enm);
 }
 
@@ -266,6 +267,7 @@ static void ctxErrResponse( BinRequestContext * ctx, CMPIStatus *rc)
    switch (ctx->rc) {
    case MSG_X_NOT_SUPPORTED:
       m = "Operation not supported yy";
+      r = CMPI_RC_ERR_NOT_SUPPORTED;
       break;
    case MSG_X_INVALID_CLASS:
       m = "Class not found";
@@ -1335,6 +1337,7 @@ static CMPIConstClass * getClass(
       }
       free(sreq);
       if (rc) CIMCSetStatusWithChars(rc, resp->rc, (char*)resp->object[0].data);
+      free(resp);
       _SFCB_RETURN(NULL);
    }
    else ctxErrResponse(&binCtx,rc);
@@ -1514,10 +1517,17 @@ extern int getControlChars(char *, char **);
 extern int setupControl(char *fn);
 extern int errno;
 
+static int localConnectCount = 0;
+static pthread_mutex_t lcc_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+#define CONNECT_LOCK()   pthread_mutex_lock(&lcc_mutex)
+#define CONNECT_UNLOCK() pthread_mutex_unlock(&lcc_mutex)
+#define CONNECT_UNLOCK_RETURN(rc) { pthread_mutex_unlock(&lcc_mutex); return (rc);}
+
 int localConnect(ClientEnv* ce, CMPIStatus *st) 
 {
    static struct sockaddr_un serverAddr;
-   int sock,rc,sfcbSocket;
+   int sock,rc=0,sfcbSocket;
    void *idData;
    unsigned long int l;
    char *user;
@@ -1531,59 +1541,84 @@ int localConnect(ClientEnv* ce, CMPIStatus *st)
       char id[64];
    } msg;
    
-   if ((sock=socket(PF_UNIX, SOCK_STREAM, 0))<0) {
-      return -1;
-      if (st) {
+   CONNECT_LOCK();  
+   if (localConnectCount == 0) {
+     if ((sock=socket(PF_UNIX, SOCK_STREAM, 0))<0) {
+       if (st) {
          st->rc=CMPI_RC_ERR_FAILED;
          st->msg=ce->ft->newString(ce,strerror(errno),NULL);
-      }
-      return -1;
-   }
-   
-   if (socketName==NULL) {
-      setupControl(NULL);
-      rc=getControlChars("localSocketPath", &socketName);    
-      sunsetControl();
-      if (rc) {
+       }
+       CONNECT_UNLOCK_RETURN(-1);
+     }
+     
+     if (socketName==NULL) {
+       setupControl(NULL);
+       rc=getControlChars("localSocketPath", &socketName);    
+       sunsetControl();
+       if (rc) {
+	 if (st) {
+	   st->rc=CMPI_RC_ERR_FAILED;
+	   st->msg=ce->ft->newString(ce,"failed to open sfcb local socket",NULL);
+	 }
          fprintf(stderr,"--- Failed to open sfcb local socket (%d)\n",rc);
-         return -2;
-      }
-   }   
-
-   serverAddr.sun_family=AF_UNIX;
-   strcpy(serverAddr.sun_path,socketName);
-   
-   if (connect(sock,(struct sockaddr*)&serverAddr,
-      sizeof(serverAddr.sun_family)+strlen(serverAddr.sun_path))<0) {
-      if (st) {
+	 close(sock);
+	 CONNECT_UNLOCK_RETURN(-2);
+       }
+     }   
+     
+     serverAddr.sun_family=AF_UNIX;
+     strcpy(serverAddr.sun_path,socketName);
+     
+     if (connect(sock,(struct sockaddr*)&serverAddr,
+		 sizeof(serverAddr.sun_family)+strlen(serverAddr.sun_path))<0) {
+       if (st) {
          st->rc=CMPI_RC_ERR_FAILED;
          st->msg=ce->ft->newString(ce,strerror(errno),NULL);
-      }
-      return -1;
+       }
+       close(sock);
+       CONNECT_UNLOCK_RETURN(-1);
+     }
+   
+     msg.size=sizeof(msg)-sizeof(msg.size);
+     msg.oper=1;
+     msg.pid=getpid();
+     user=getenv("USER");
+     strncpy(msg.id, (user) ? user : "", sizeof(msg.id)-1);
+     msg.id[sizeof(msg.id)-1]=0;
+     
+     l=write(sock, &msg, sizeof(msg)); 
+     
+     rc = spRecvCtlResult(&sock, &sfcbSocket,&idData, &l);
+     if (rc < 0 || sfcbSocket <= 0) {
+       if (st) {
+         st->rc=CMPI_RC_ERR_FAILED;
+         st->msg=ce->ft->newString(ce,"failed to get socket fd for local connect",NULL);
+       }
+       fprintf(stderr,"--- Failed to get socket fd for local connect (%d, %d, %lu)\n", rc, sfcbSocket, l);
+       close(sock);
+       CONNECT_UNLOCK_RETURN(-3);
+     }
+     sfcbSockets.send=sfcbSocket;
+
+     close(sock);
    }
-   
-   msg.size=sizeof(msg)-sizeof(msg.size);
-   msg.oper=1;
-   msg.pid=getpid();
-   user=getenv("USER");
-   strncpy(msg.id, (user) ? user : "", sizeof(msg.id)-1);
-   msg.id[sizeof(msg.id)-1]=0;
-   
-   l=write(sock, &msg, sizeof(msg)); 
-   
-   rc = spRecvCtlResult(&sock, &sfcbSocket,&idData, &l);
-  
-   sfcbSockets.send=sfcbSocket;
+   localConnectCount += 1;
+   CONNECT_UNLOCK();
    localMode=0;
-   close(sock);
    
-   return sfcbSocket;
+   return (rc == 0) ? rc : sfcbSocket;
 }
 
 static void* release(ClientEnv* ce)
 {
    void *lib=ce->hdl;
-   close(sfcbSockets.send);
+   CONNECT_LOCK();
+   if (localConnectCount > 0) localConnectCount -= 1;
+   if (localConnectCount == 0) {
+     close(sfcbSockets.send);
+     sfcbSockets.send = -1;
+   }
+   CONNECT_UNLOCK();
    free(ce);
    uninitGarbageCollector();
    return lib;
