@@ -24,15 +24,19 @@
 #include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
+#include <dlfcn.h>
 
 #include "cimslpCMPI.h"
 #include "cimslpSLP.h"
 #include "cimslpUtil.h"
 #include "trace.h"
-
+#include "config.h"
+#include "control.h"
 
 char * interOpNS;
-
+typedef int (*getSlpHostname)(char **hostname);
+extern void libraryName(const char *dir, const char *location, const char *fullName, int buf_size);
+extern char * configfile;
 
 //helper function ... until better solution is found to get to the
 //interop Namespace
@@ -68,6 +72,9 @@ CMPIInstance ** myGetInstances(CMCIClient *cc, char * path, char * objectname)
                 //printf("Fatal error. CIMOM not running? Connection params ok?\n");
                 printf("--- HTTP-Daemon no longer active. Deregistering service with slp\n");
                 deregisterCIMService();
+                if (status.msg) CMRelease(status.msg);  
+                if (objectpath) CMRelease(objectpath);
+                if (enumeration) CMRelease(enumeration);                                
                 exit(0);
         }
         
@@ -127,15 +134,36 @@ cimSLPService getSLPData(cimomConfig cfg)
         CMPIInstance **ci;
         CMPIStatus status;
         CMPIConstClass *ccls;
-
         cimSLPService rs; //service which is going to be returned to the calling function
+        char *sn;
+
+#ifdef SLP_HOSTNAME_LIB
+        static void *hostnameLib=NULL;
+        static getSlpHostname gethostname; 
+        char *ln;
+        char dlName[512];
+        int err;
+        
+        err = 1;
+        setupControl(configfile);
+        if (getControlChars("slpHostnamelib", &ln) == 0) {
+           libraryName(NULL,ln,dlName, 512);
+           if ((hostnameLib = dlopen(dlName, RTLD_LAZY))) {
+              gethostname = dlsym(hostnameLib, "_sfcGetSlpHostname");
+              if (gethostname) err = 0;
+           }
+        }
+        sunsetControl();
+        if (err) mlogf(M_ERROR,M_SHOW,"--- SLP Hostname exit %s not found. Defaulting to system hostname.\n",
+              dlName);
+#endif
         
         _SFCB_ENTER(TRACE_SLP, "getSLPData");
         
         memset(&rs, 0, sizeof(cimSLPService));
 
         cc = cmciConnect2(cfg.cimhost,
-                                        cfg.commScheme,
+                                        "SfcbLocal",
                                         cfg.port,
                                         cfg.cimuser,
                                         cfg.cimpassword,
@@ -160,8 +188,21 @@ cimSLPService getSLPData(cimomConfig cfg)
         //construct the server string
         ci = myGetInstances(cc, interOpNS, "CIM_ObjectManager");
         if(ci) {
-                rs.url_syntax = getUrlSyntax(myGetProperty(ci[0], "SystemName"),
-                                                                        cfg.commScheme, cfg.port);      
+                sn = myGetProperty(ci[0], "SystemName");
+
+#ifdef SLP_HOSTNAME_LIB
+                if (!err) {
+                  char *tmp;
+                  if ((err = gethostname(&tmp))) {
+                     free(sn);
+                     sn = tmp;
+                  } else {
+                     printf("-#- SLP call to %s for hostname failed. Defaulting to system hostname.\n", dlName);
+                  }
+                }
+#endif
+                rs.url_syntax = getUrlSyntax(sn, cfg.commScheme,
+                      cfg.port);
                 rs.service_hi_name = myGetProperty(ci[0], "ElementName");
                 rs.service_hi_description = myGetProperty(ci[0], "Description");
                 rs.service_id = myGetProperty(ci[0], "Name");
@@ -267,6 +308,7 @@ char ** myGetRegProfiles(CMPIInstance **instances, CMCIClient *cc)
     objectpath = instances[i]->ft->getObjectPath(instances[i], &status);
     if(status.rc) {
       //no object path ??
+      free(retArr);
       _SFCB_RETURN(NULL);
     }
     objectpath->ft->setNameSpace(objectpath, interOpNS);
@@ -385,7 +427,10 @@ char * transformValue(char* cssf, CMPIConstClass * ccls, char * propertyName)
                         free(valuestr);
                         ele = CMGetArrayElementAt(arr, j, NULL);
                         valuestr = value2Chars(eletyp, &ele.value);
-                        if(j == n) _SFCB_RETURN(cssf); //nothing found, probably "NULL" -> return it
+                        if(j == n) {
+                             free(valuestr);
+                             _SFCB_RETURN(cssf); //nothing found, probably "NULL" -> return it
+                        } 
                         j++;
                 }
                 free(valuestr);
