@@ -124,6 +124,7 @@ typedef struct nameSpaces {
 } NameSpaces;
 
 static UtilHashTable *nsHt=NULL;
+static pthread_once_t nsHt_once = PTHREAD_ONCE_INIT;
 static int nsBaseLen;
 
 static void buildInheritanceTable(ClassRegister * cr)
@@ -402,7 +403,7 @@ static ClassRegister *newClassRegister(char *fname)
       free(buf);
       free(cc);
    }
- 
+
    if (cr->vr) {
       mlogf(M_INFO,M_SHOW,"--- Caching ClassProvider for %s (%d.%d-%d) using %ld bytes\n", 
           fin, cr->vr->version, cr->vr->level, cr->vr->objImplLevel, total);
@@ -445,11 +446,10 @@ static UtilHashTable *gatherNameSpaces(char *dn, UtilHashTable *ns, int first)
      closedir(dir_test);
      cr=newClassRegister(n);
      if (cr) {
-       ns->ft->put(ns, n+nsBaseLen, cr);
+       ns->ft->put(ns, strdup(n+nsBaseLen), cr);  
        gatherNameSpaces(n,ns,0);
-     } else {
-        free(n);
      }
+     free(n);
    }
    else if (first) {
       mlogf(M_ERROR,M_SHOW,"--- Repository %s not found\n",dn);
@@ -476,6 +476,11 @@ static UtilHashTable *buildClassRegisters()
    return gatherNameSpaces(dn,NULL,1);   
 }    
 
+static void nsHt_init()
+{
+  nsHt=buildClassRegisters();
+}
+
 
 static ClassRegister *getNsReg(const CMPIObjectPath *ref, int *rc)
 {
@@ -483,9 +488,15 @@ static ClassRegister *getNsReg(const CMPIObjectPath *ref, int *rc)
    CMPIString *nsi=CMGetNameSpace(ref,NULL);
    ClassRegister *cReg;
    *rc=0;
-   
-   if (nsHt==NULL) nsHt=buildClassRegisters();
-   
+
+   pthread_once(&nsHt_once, nsHt_init);
+
+   if (nsHt==NULL) {
+     mlogf(M_ERROR,M_SHOW,"--- ClassProvider: namespace hash table not initialized\n");
+     *rc = 1;
+     return NULL;
+   }
+
    if (nsi && nsi->hdl) {
       ns=(char*)nsi->hdl;
       if (strcasecmp(ns,"root/pg_interop")==0)
@@ -643,7 +654,10 @@ static CMPIStatus ClassProviderEnumClassNames(CMPIClassMI * mi,
    ns=(char*)CMGetNameSpace(ref,NULL)->hdl;
    flgs=ctx->ft->getEntry(ctx,CMPIInvocationFlags,NULL).value.uint32;
    cni=ref->ft->getClassName(ref,NULL);
-   if (cni) cn=(char*)cni->hdl;
+   if (cni) {
+      cn=(char*)cni->hdl;
+      if (cn && *cn==0) cn=NULL;
+   }   
    cb = (ClassBase *) cReg->hdl;  
 
    cReg->ft->wLock(cReg);
@@ -775,7 +789,7 @@ static CMPIStatus ClassProviderGetClass(CMPIClassMI * mi,
 {
    CMPIStatus st = { CMPI_RC_OK, NULL };
    CMPIString *cn = CMGetClassName(ref, NULL);
-   CMPIConstClass *cl;
+   CMPIConstClass *cl, *clLocal;
    ClassRegister *cReg;
    int rc;
    void *cid;
@@ -791,9 +805,14 @@ static CMPIStatus ClassProviderGetClass(CMPIClassMI * mi,
 
    cReg->ft->wLock(cReg);
    
-   cl = getClass(cReg, (char *) cn->hdl,&cid);
-   if (cl) {
+   /* Make a cloned copy of the cached results to prevent
+      thread interference. */
+   clLocal = getClass(cReg, (char *) cn->hdl,&cid);
+
+   if (clLocal) {
       _SFCB_TRACE(1,("--- Class found"));
+      cl = clLocal->ft->clone(clLocal, NULL);
+      memLinkInstance((CMPIInstance*)cl);
       CMReturnInstance(rslt, (CMPIInstance *) cl);
    } else {
       _SFCB_TRACE(1,("--- Class not found"));
