@@ -54,14 +54,29 @@
 #include <sys/ipc.h>
 #include <sys/sem.h>
 
+#ifdef HAVE_SYS_STAT_H
+#include <sys/stat.h>
+#endif
+
+#ifdef HAVE_SYS_FSUID_H
+#include <sys/fsuid.h>
+#endif
+
 #include "httpComm.h"
 #include "sfcVersion.h"
 #include "control.h"
 
+#ifdef HAVE_UDS
+#include <grp.h>
+#endif
+
 unsigned long exFlags = 0;
 static char *name;
 static int debug;
 static int doBa;
+#ifdef HAVE_UDS
+static int doUdsAuth;
+#endif
 static int doFork = 0;
 int noChunking = 0;
 int sfcbSSLMode = 0;
@@ -74,6 +89,7 @@ static int running=0;
 static long keepaliveTimeout=15;
 static long keepaliveMaxRequest=10;
 static long numRequest;
+struct timeval httpSelectTimeout = {5, 0};  /* 5 sec. timeout for select() before read() */
 
 #if defined USE_SSL
 static SSL_CTX *ctx;
@@ -92,7 +108,7 @@ static int httpProcSem;
 static int httpWorkSem;
 
 extern char *decode64(char *data);
-extern void libraryName(const char* dir, const char *location, char *fullName);
+extern void libraryName(const char* dir, const char *location, char *fullName, int buf_size);
 extern void *loadLibib(const char *libname);
 extern int getControlChars(char *id, char **val);
 extern void *loadLibib(const char *libname);
@@ -114,7 +130,9 @@ typedef int (*Authenticate)(char* principal, char* pwd);
 
 typedef struct _buffer {
    char *data, *content;
-   int length, size, ptr, content_length,trailers;
+   int length, size, ptr;
+   unsigned int content_length;
+   int trailers;
    char *httpHdr, *authorization, *content_type, *host, *useragent;
    char *principal;
    char *protocol;
@@ -160,7 +178,7 @@ void initHttpProcCtl(int p, int sslmode)
    
    sun.val=0;
    for (i=1; i<=p; i++)
-      semctl(httpProcSem,p,SETVAL,sun);
+      semctl(httpProcSem,i,SETVAL,sun);
 
    if ((httpWorkSem=semget(httpWorkSemKey,1,0600))!=-1) 
       semctl(httpWorkSem,0,IPC_RMID,sun);
@@ -203,9 +221,9 @@ int baValidate(char *cred, char **principal)
 
    if (err==0 && authLib==NULL) {
       char *ln;
-      int err=1;
+      err=1;
       if (getControlChars("basicAuthlib", &ln)==0) {
-         libraryName(NULL,ln,dlName);
+         libraryName(NULL,ln,dlName, 512);
         if ((authLib=dlopen(dlName, RTLD_LAZY))) {
             authenticate= dlsym(authLib, "_sfcBasicAuthenticate");
             if (authenticate) err=0;
@@ -214,9 +232,14 @@ int baValidate(char *cred, char **principal)
       if (err) mlogf(M_ERROR,M_SHOW,"--- Authentication exit %s not found\n",dlName);
    }
 
-   *principal=strdup(auth);
-   if (authenticate(auth,pw)) 
-   		ret = 1;
+   if (err) {
+     ret = 1;
+   } 
+   else {
+     *principal=strdup(auth);
+     if (authenticate(auth,pw)) 
+       ret = 1;
+   }
 
    free(auth);
 
@@ -251,18 +274,40 @@ static void handleSigChld(int sig)
    errno = oerrno;
 }
 
+static void stopProc(void *p)
+{
+//   printf("--- %s draining %d\n",processName,running);
+   for (;;) {
+      if (running==0) {
+         mlogf(M_INFO,M_SHOW,"--- %s terminating %d\n",processName,getpid());
+         exit(0);
+      }   
+      sleep(1);
+   }   
+}
+
 static void handleSigUsr1(int sig)
 {
-   stopAccepting=1;
+   pthread_t t;
+   pthread_attr_t tattr;
+
+   if (stopAccepting == 0) {
+     stopAccepting=1;
+     pthread_attr_init(&tattr);
+     pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);      
+     pthread_create(&t, &tattr, (void *(*)(void *))stopProc,NULL);
+   }
 }
 
 static void freeBuffer(Buffer * b)
 {
-   Buffer emptyBuf = { NULL, NULL, 0, 0, 0, 0, 0 ,0};
+   Buffer emptyBuf = { NULL, NULL, 0, 0, 0, 0, 0 , NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    if (b->data)
       free(b->data);
    if (b->content)
       free(b->content);
+   if (b->principal)
+      free(b->principal);
    *b=emptyBuf;   
 }
 
@@ -333,26 +378,47 @@ static void genError(CommHndl conn_fd, Buffer * b, int status, char *title,
 
 static int readData(CommHndl conn_fd, char *into, int length)
 {
-   int c = 0, r;
+   int c = 0, r, isReady;
+   fd_set httpfds;
+   FD_ZERO(&httpfds);
+   FD_SET(conn_fd.socket,&httpfds);
 
    while (c < length) {
+      isReady = select(conn_fd.socket+1,&httpfds,NULL,NULL,&httpSelectTimeout);
+      if (isReady == 0) {
+         c = -1;
+         break;
+      }
       r = commRead(conn_fd, into + c, length - c);
       if (r < 0 && (errno == EINTR || errno == EAGAIN)) {
          continue;
       }
+      /* r==0 is a success condition for read(), but the loop should complete prior to this */
+      else if (r == 0) {
+         mlogf(M_INFO,M_SHOW,"--- commRead hit EOF sooner than expected\n");
+         c = -2;
+	 break;
+      }
       c += r;
    }
    return c;
 }
 
-static void getPayload(CommHndl conn_fd, Buffer * b)
+static int getPayload(CommHndl conn_fd, Buffer * b)
 {
-   int c = b->length - b->ptr;
+   unsigned int c = b->length - b->ptr;
+   int rc = 0;
    b->content = (char *) malloc(b->content_length + 8);
    if (c) memcpy(b->content, (b->data) + b->ptr, c);
 
-   readData(conn_fd, (b->content) + c, b->content_length - c);
+   if (c > b->content_length) {
+     mlogf(M_INFO,M_SHOW,"--- HTTP Content-Length is lying; content truncated\n");
+     c = b->content_length;
+   }
+
+   rc = readData(conn_fd, (b->content) + c, b->content_length - c);
    *((b->content) + b->content_length) = 0;
+   return rc;
 }
 
 void dumpResponse(RespSegments * rs)
@@ -428,6 +494,7 @@ static void writeResponse(CommHndl conn_fd, RespSegments rs)
          }
       }
    }
+
    commFlush(conn_fd);
    
    _SFCB_EXIT();
@@ -571,30 +638,37 @@ static ChunkFunctions httpChunkFunctions = {
 #define hdrBufsize 5000
 #define hdrLimmit 5000
 
-static int  getHdrs(CommHndl conn_fd, Buffer * b, char *cmd)
+static int getHdrs(CommHndl conn_fd, Buffer * b, char *cmd)
 {
    int first=1,total=0,isReady;
-   struct timeval httpTimeout;
    fd_set httpfds;
    int state=0;
    
    FD_ZERO(&httpfds);
    FD_SET(conn_fd.socket,&httpfds);
-   httpTimeout.tv_sec=5;
-   httpTimeout.tv_usec=0;
-   isReady = select(conn_fd.socket+1,&httpfds,NULL,NULL,&httpTimeout);
-   if (isReady == 0) return 3;
 	
    for (;;) {
+      isReady = select(conn_fd.socket+1,&httpfds,NULL,NULL,&httpSelectTimeout);
+      if (isReady == 0) return 3;
+
       char buf[hdrBufsize];
       int r = commRead(conn_fd, buf, sizeof(buf));
       
       if (r < 0 && (errno == EINTR || errno == EAGAIN)) continue;
-      if (r <= 0) break;
+      if (r == 0) {
+        if (b->size == 0) break;
+        if (strstr(b->data, "\r\n\r\n") == NULL &&
+            strstr(b->data, "\n\n") == NULL) {
+	        mlogf(M_ERROR,M_SHOW,"-#- HTTP header ended prematurely\n");
+      	    state = 3;
+	        break; 
+	    }
+      }
       
       add2buffer(b, buf, r);
       total+=r;
-//      fprintf(stderr,"+++ buf: >%s<\n",buf);
+
+      /* on first run through, ensure that this is a POST req. */
       if (r && first) {
          if (strncasecmp(buf,cmd,strlen(cmd)) != 0) { 
 	   /* not what we expected - still continue to read to
@@ -603,7 +677,8 @@ static int  getHdrs(CommHndl conn_fd, Buffer * b, char *cmd)
 	 }
          first=0;
       }
-      
+
+      /* success condition: end of header */
       if (strstr(b->data, "\r\n\r\n") != NULL ||
           strstr(b->data, "\n\n") != NULL) {
          break;
@@ -652,7 +727,8 @@ static int doHttpRequest(CommHndl conn_fd)
    Buffer inBuf = { NULL, NULL, 0, 0, 0, 0, 0 ,0};
    RespSegments response;
    static RespSegments nullResponse = { NULL, 0, 0, NULL, { {0, NULL} } };
-   int len, hl, rc,uset=0;
+   unsigned long len;
+   int hl, rc,uset=0;
    char *hdr, *path;
    int discardInput=0;
    MsgSegment msgs[2];
@@ -671,13 +747,13 @@ static int doHttpRequest(CommHndl conn_fd)
    inBuf.authorization = "";
    inBuf.protocol="HTTP/1.1";
    inBuf.content_type = NULL;
-   inBuf.content_length = -1;
+   inBuf.content_length = UINT_MAX;
    inBuf.host = NULL;
    inBuf.useragent = "";
    int badReq = 0;
 
    rc=getHdrs(conn_fd, &inBuf,"POST ");
-   
+
    if (rc==1) { 
       genError(conn_fd, &inBuf, 501, "Not Implemented", NULL);
       /* we continue to parse headers and empty the socket
@@ -743,7 +819,34 @@ static int doHttpRequest(CommHndl conn_fd)
       else if (strncasecmp(hdr, "Content-Length:", 15) == 0) {
          cp = &hdr[15];
          cp += strspn(cp, " \t");
-         inBuf.content_length = atol(cp);
+         if (cp[0] == '-') {
+           genError(conn_fd, &inBuf, 400, "Negative Content-Length", NULL);
+           _SFCB_TRACE(1, ("--- exiting: content-length too big"));
+           commClose(conn_fd);
+           exit(1);
+         }
+         errno = 0; 
+         unsigned long clen = strtoul(cp, NULL, 10);
+         if (errno != 0) {
+           genError(conn_fd, &inBuf, 400, "Error converting Content-Length to a decimal value", NULL);
+           _SFCB_TRACE(1, ("--- exiting: content-length conversion error"));
+           commClose(conn_fd);
+           exit(1);
+         }
+         unsigned int maxLen;
+         if (getControlUNum("httpMaxContentLength", &maxLen) != 0) {
+           genError(conn_fd, &inBuf, 501, "Server misconfigured (httpMaxContentLength)", NULL);
+           _SFCB_TRACE(1, ("--- exiting: bad config httpMaxContentLength"));
+           commClose(conn_fd);
+           exit(1);
+         }
+         if((clen >= UINT_MAX) || ((maxLen) && (clen > maxLen))) {
+            genError(conn_fd, &inBuf, 413, "Request Entity Too Large", NULL);
+            _SFCB_TRACE(1, ("--- exiting: content-length too big"));      
+            commClose(conn_fd);
+            exit(1);
+         }
+         inBuf.content_length = clen;
       }
       else if (strncasecmp(hdr, "Content-Type:", 13) == 0) {
          cp = &hdr[13];
@@ -800,7 +903,20 @@ static int doHttpRequest(CommHndl conn_fd)
    }
 #endif
 
-   if (!discardInput && doBa) {
+   int authorized = 0; 
+#ifdef HAVE_UDS
+   if (!discardInput && doUdsAuth) {
+       struct sockaddr_un sun; 
+       sun.sun_family = 0; 
+       socklen_t cl = sizeof(sun); 
+       int rc = getpeername(conn_fd.socket, (struct sockaddr*)&sun, &cl); 
+       if (rc == 0 && sun.sun_family == AF_UNIX) {
+           /* Already authenticated via permissions on unix socket */
+           authorized = 1;
+       }
+   }
+#endif
+   if (!authorized && !discardInput && doBa) {
      if (!(inBuf.authorization && baValidate(inBuf.authorization,&inBuf.principal))) {
        char more[]="WWW-Authenticate: Basic realm=\"cimom\"\r\n";
        genError(conn_fd, &inBuf, 401, "Unauthorized", more);
@@ -817,12 +933,13 @@ static int doHttpRequest(CommHndl conn_fd)
    }
 
    len = inBuf.content_length;
-   if (len < 0) {
+   if (len == 0) {
      if (!discardInput) {
         genError(conn_fd, &inBuf, 411, "Length Required", NULL);
      }
      _SFCB_TRACE(1, ("--- exiting after missing content length."));      
      commClose(conn_fd);
+     freeBuffer(&inBuf);
      exit(1);
    }
 
@@ -830,7 +947,13 @@ static int doHttpRequest(CommHndl conn_fd)
    len += hl =
        sprintf(hdr, "<!-- xml -->\n<!-- auth: %s -->\n", inBuf.authorization);
 
-   getPayload(conn_fd, &inBuf);
+   rc = getPayload(conn_fd, &inBuf);
+   if (rc < 0) {
+      genError(conn_fd, &inBuf, 400, "Bad Request", NULL);
+      _SFCB_TRACE(1, ("--- exiting after request timeout."));
+      commClose(conn_fd);
+      exit(1);
+   }
    if (discardInput) {
      free(hdr);
      freeBuffer(&inBuf);
@@ -1272,12 +1395,10 @@ static void handleHttpRequest(int connFd)
 	    /* still in handshake */
 	    FD_ZERO(&httpfds);
 	    FD_SET(connFd,&httpfds);
-	    httpTimeout.tv_sec=5;
-	    httpTimeout.tv_usec=0;
 	    if (sslerr == SSL_ERROR_WANT_WRITE) {
-	      isReady = select(connFd+1,NULL,&httpfds,NULL,&httpTimeout);
+	      isReady = select(connFd+1,NULL,&httpfds,NULL,&httpSelectTimeout);
 	    } else {
-	      isReady = select(connFd+1,&httpfds,NULL,NULL,&httpTimeout);
+	      isReady = select(connFd+1,&httpfds,NULL,NULL,&httpSelectTimeout);
 	    }
 	    if (isReady == 0) {
 	      intSSLerror("Timeout error accepting SSL connection");
@@ -1346,6 +1467,19 @@ static void handleHttpRequest(int connFd)
 
 }
 
+int isDir(const char *path)
+{
+#if defined (HAVE_SYS_STAT_H) && defined (USE_SSL)
+   struct stat sb;
+   int rc;
+   rc = stat(path, &sb);
+   if (rc == -1)
+      intSSLerror("Error opening the client trust store");
+   return (sb.st_mode & S_IFMT) == S_IFDIR;
+#else
+   return 0;
+#endif
+}
 
 int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
 {
@@ -1355,12 +1489,22 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
 #else
    struct sockaddr_in sin;
 #endif
+   struct sockaddr_un sun; 
 
-   socklen_t sz,sin_len;
-   int i,ru;
+   socklen_t sz,sin_len,sun_len;
+   int i,ru,rc;
    char *cp;
    long procs, port;
-   int listenFd, connFd;
+   int listenFd=-1, connFd; 
+   int enableHttp=0;
+   fd_set httpfds;
+   int maxfdp1; 
+
+#ifdef HAVE_UDS
+   static char *udsPath=NULL;
+   int enableUds=0;
+   int  udsListenFd=-1;
+#endif
 
    name = argv[0];
    debug = 1;
@@ -1382,6 +1526,10 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
    else {
       if (getControlNum("httpPort", &port))
          port = 5988;
+#ifdef HAVE_UDS
+      if (getControlChars("httpSocketPath", &udsPath)) 
+		 udsPath = "/tmp/sfcbHttpSocket"; 
+#endif
       hBase=htBase;
       hMax=htMax;
    }
@@ -1392,12 +1540,27 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
    } else {
      if (getControlNum("httpProcs", &procs))
        procs = 10;
+     if (getControlBool("enableHttp", &enableHttp))
+       enableHttp=1;
+#ifdef HAVE_UDS
+	 if (getControlBool("enableUds", &enableUds))
+       enableUds=1;
+	 if (!enableUds)
+		 udsPath = NULL; 
+#endif
+	 if (!enableHttp)
+		 port = -1; 
    }
    initHttpProcCtl(procs,sslMode);
 
    if (getControlBool("doBasicAuth", &doBa))
       doBa=0;
 
+#ifdef HAVE_UDS
+   if (getControlBool("doUdsAuth", &doUdsAuth))
+      doUdsAuth=0;
+#endif
+
    if (getControlNum("keepaliveTimeout", &keepaliveTimeout))
      keepaliveTimeout = 15;
 
@@ -1439,11 +1602,15 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
 
    if (sslMode) mlogf(M_INFO,M_SHOW,"--- %s HTTPS Daemon V" sfcHttpDaemonVersion " started - %d - port %ld\n", 
          name, currentProc,port);
-   else mlogf(M_INFO,M_SHOW,"--- %s HTTP  Daemon V" sfcHttpDaemonVersion " started - %d - port %ld\n", 
-         name, currentProc,port);
-
+#ifdef HAVE_UDS
+   else mlogf(M_INFO,M_SHOW,"--- %s HTTP  Daemon V" sfcHttpDaemonVersion " started - %d - port %ld, %s\n", 
+         name, currentProc,port,udsPath);
+#endif
 
    if (doBa) mlogf(M_INFO,M_SHOW,"--- Using Basic Authentication\n");
+#ifdef HAVE_UDS
+   if (doUdsAuth) mlogf(M_INFO,M_SHOW,"--- Using Unix Socket Peer Cred Authentication\n");
+#endif
 
    if (keepaliveTimeout == 0) {
      mlogf(M_INFO,M_SHOW,"--- Keep-alive timeout disabled\n");
@@ -1452,51 +1619,112 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
      mlogf(M_INFO,M_SHOW,"--- Maximum requests per connection: %ld\n",keepaliveMaxRequest);
    }
 
-#ifdef USE_INET6
-   listenFd = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
-   if (listenFd < 0) { 
-       mlogf(M_INFO,M_SHOW,"--- Using IPv4 address\n");
-       listenFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+   ru = 1;
+#ifdef HAVE_UDS
+   if (enableUds) {
+      udsListenFd = socket(PF_UNIX, SOCK_STREAM, 0); 
    }
-#else
-   listenFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
 #endif
+   if (enableHttp || sslMode) {
+#ifdef USE_INET6
+      listenFd = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
+      if (listenFd < 0) { 
+          mlogf(M_INFO,M_SHOW,"--- Using IPv4 address\n");
+          listenFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+      }
+#else
+      listenFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+#endif
+      setsockopt(listenFd, SOL_SOCKET, SO_REUSEADDR, (char *) &ru, sizeof(ru));
+   }
 
    sin_len = sizeof(sin);
-
-   ru = 1;
-   setsockopt(listenFd, SOL_SOCKET, SO_REUSEADDR, (char *) &ru, sizeof(ru));
+   sun_len = sizeof(sun); 
 
    memset(&sin,0,sin_len);
+   memset(&sun,0,sun_len);
 
-   if (getControlBool("httpLocalOnly", &httpLocalOnly))
-      httpLocalOnly=0;
+#ifdef HAVE_UDS
+   if (udsListenFd >= 0) {
+     if (getControlChars("httpSocketPath", &udsPath)) {
+         mlogf(M_ERROR,M_SHOW,"--- No unix socket path defined for HTTP\n"); 
+         sleep(1);
+         kill(sfcbPid,3);
+     }
+     sun.sun_family=AF_UNIX;
+     strcpy(sun.sun_path,udsPath);
+   }
+#endif
+
+   if (listenFd >= 0) {
+      if (getControlBool("httpLocalOnly", &httpLocalOnly))
+	    httpLocalOnly=0;
 
 #ifdef USE_INET6
-   sin.sin6_family = AF_INET6;
-   if (httpLocalOnly)
-     sin.sin6_addr = in6addr_loopback;
-   else
-     sin.sin6_addr = in6addr_any;
-   sin.sin6_port = htons(port);
+     sin.sin6_family = AF_INET6;
+     if (httpLocalOnly)
+	   sin.sin6_addr = in6addr_loopback;
+     else
+	   sin.sin6_addr = in6addr_any;
+     sin.sin6_port = htons(port);
 #else
-   sin.sin_family = AF_INET;
-   if (httpLocalOnly) {
-     char* loopback_int = "127.0.0.1";
-     inet_aton(loopback_int, &sin.sin_addr); /* not INADDR_LOOPBACK ? */
-   }
-   else
-     sin.sin_addr.s_addr = INADDR_ANY;
-   sin.sin_port = htons(port);
+     sin.sin_family = AF_INET;
+     if (httpLocalOnly) {
+	   char* loopback_int = "127.0.0.1";
+	   inet_aton(loopback_int, &sin.sin_addr); /* not INADDR_LOOPBACK ? */
+     }
+     else
+	   sin.sin_addr.s_addr = INADDR_ANY;
+     sin.sin_port = htons(port);
 #endif 
-
-   if (bind(listenFd, (struct sockaddr *) &sin, sin_len) ||
-       listen(listenFd, 0)) {
-      mlogf(M_ERROR,M_SHOW,"--- Cannot listen on port %ld (%s)\n", port, strerror(errno));
-      sleep(1);
-      kill(sfcbPid,3);
    }
 
+  if (listenFd >= 0) {
+     if (bind(listenFd, (struct sockaddr *) &sin, sin_len) ||
+             listen(listenFd, 10)) {
+            mlogf(M_ERROR,M_SHOW,"--- Cannot listen on port %ld (%s)\n", port, strerror(errno));
+            sleep(1);
+            kill(sfcbPid,3);
+     }
+  }
+#ifdef HAVE_UDS
+  if (udsListenFd >= 0) {
+     unlink(udsPath); 
+
+     size_t gbuflen = sysconf(_SC_GETGR_R_SIZE_MAX); 
+     char gbuf[gbuflen]; 
+     struct group* pgrp = NULL; 
+     struct group grp; 
+     gid_t oldfsgid = 0; 
+
+     int rc = getgrnam_r("sfcb", &grp, gbuf, gbuflen, &pgrp); 
+     if (rc == 0 && pgrp)
+     {
+#ifdef HAVE_SYS_FSUID_H
+         oldfsgid = setfsgid(pgrp->gr_gid); 
+#else
+         oldfsgid = setegid(pgrp->gr_gid); 
+#endif
+     }
+     mode_t oldmask = umask(0007); 
+     if (bind(udsListenFd, (struct sockaddr *) &sun, sun_len) ||
+             listen(udsListenFd, 10)) {
+            mlogf(M_ERROR,M_SHOW,"--- Cannot listen on unix socket %s (%s)\n", udsPath, strerror(errno));
+            sleep(1);
+            kill(sfcbPid,3);
+     }
+     umask(oldmask); 
+     if (pgrp)
+     {
+#ifdef HAVE_SYS_FSUID_H
+         setfsgid(oldfsgid); 
+#else
+         setegid(oldfsgid); 
+#endif
+     }
+  }
+#endif
+
   if (!debug) {
       int rc = fork();
       if (rc == -1) {
@@ -1524,6 +1752,7 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
 #if defined USE_SSL
     if (sfcbSSLMode) {
        char *fnc,*fnk, *fnt, *fnl;
+       int rc;
        ctx = SSL_CTX_new(SSLv23_method());
        getControlChars("sslCertificateFilePath", &fnc);
        _SFCB_TRACE(1,("---  sslCertificateFilePath = %s",fnc));
@@ -1551,43 +1780,89 @@ int httpDaemon(int argc, char *argv[], int sslMode, int sfcbPid)
        } 
        getControlChars("sslClientTrustStore", &fnt);
        _SFCB_TRACE(1,("---  sslClientTrustStore = %s",fnt));
-       if (ccVerifyMode != CC_VERIFY_IGNORE &&
-	   SSL_CTX_load_verify_locations(ctx, fnt, NULL) != 1)
-	 intSSLerror("Error locating the client trust store");
+
+       if (ccVerifyMode != CC_VERIFY_IGNORE) {
+           if (isDir(fnt))
+              rc = SSL_CTX_load_verify_locations(ctx, NULL, fnt);
+           else
+              rc = SSL_CTX_load_verify_locations(ctx, fnt, NULL);
+           if (rc != 1)
+              intSSLerror("Error locating the client trust store");
+       }
+
+       /* SSLv2 is pretty old; no one should be needing it any more */
+       SSL_CTX_set_options(ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2 | 
+                           SSL_OP_SINGLE_DH_USE); 
+       /* disable weak ciphers */
+       if (SSL_CTX_set_cipher_list(ctx, "ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH") != 1)
+           intSSLerror("Error setting cipher list (no valid ciphers)"); 
 	      
     }
 #endif
 
+#ifdef HAVE_UDS
+   maxfdp1 = (listenFd > udsListenFd? listenFd : udsListenFd) + 1; 
+#else
+   maxfdp1 = listenFd + 1; 
+#endif
    for (;;) {
-   char *emsg;
-      listen(listenFd, 1);
-      sz = sizeof(sin);
-      if ((connFd = accept(listenFd, (__SOCKADDR_ARG) & sin, &sz))<0) {
-         if (errno == EINTR || errno == EAGAIN) {
-            if (stopAccepting) break;
-            continue;
-         }   
-         emsg=strerror(errno);
-         mlogf(M_ERROR,M_SHOW,"--- accept error %s\n",emsg);
-         _SFCB_ABORT();
+      char *emsg;
+      // listen(listenFd, 1);
+	  FD_ZERO(&httpfds); 
+	  if (listenFd >= 0) {
+         FD_SET(listenFd, &httpfds); 
+	  }
+#ifdef HAVE_UDS
+	  if (udsListenFd >= 0) {
+         FD_SET(udsListenFd, &httpfds); 
       }
-      _SFCB_TRACE(1, ("--- Processing http request"));
+#endif
+	  rc = select(maxfdp1, &httpfds, NULL, NULL, NULL); 
+	  if (stopAccepting) break;
+      if (rc < 0) {
+         if (errno == EINTR || errno == EAGAIN) {
+            continue;
+         }
+      }
+	  if (listenFd >= 0 && FD_ISSET(listenFd, &httpfds)) {
+		  sz = sin_len; 
+		  if ((connFd = accept(listenFd, (__SOCKADDR_ARG) &sin, &sz))<0) {
+			 if (errno == EINTR || errno == EAGAIN) {
+				continue;
+			 }   
+			 emsg=strerror(errno);
+			 mlogf(M_ERROR,M_SHOW,"--- accept error %s\n",emsg);
+			 _SFCB_ABORT();
+		  }
+		  _SFCB_TRACE(1, ("--- Processing http request"));
 
-      handleHttpRequest(connFd);
-      close(connFd);
+		  handleHttpRequest(connFd);
+		  close(connFd);
+	  }
+#ifdef HAVE_UDS
+	  if (udsListenFd >= 0 && FD_ISSET(udsListenFd, &httpfds)) {
+		  sz = sun_len; 
+		  if ((connFd = accept(udsListenFd, (__SOCKADDR_ARG) &sun, &sz))<0) {
+			 if (errno == EINTR || errno == EAGAIN) {
+				continue;
+			 }   
+			 emsg=strerror(errno);
+			 mlogf(M_ERROR,M_SHOW,"--- accept error %s\n",emsg);
+			 _SFCB_ABORT();
+		  }
+		  _SFCB_TRACE(1, ("--- Processing http request"));
+
+		  handleHttpRequest(connFd);
+		  close(connFd);
+	  }
+#endif
    }
    
    remProcCtl();   
    
-//   printf("--- %s draining %d\n",processName,running);
-   for (;;) {
-      if (running==0) {
-         mlogf(M_INFO,M_SHOW,"--- %s terminating %d\n",processName,getpid());
-         exit(0);
-      }   
-      sleep(1);
-   }   
-
+   while (1) {
+     sleep(5);
+   }
 }
 
 #if defined USE_SSL
@@ -1611,7 +1886,7 @@ static int ccValidate(X509 *certificate, char ** principal, int mode)
   _SFCB_ENTER(TRACE_HTTPDAEMON, "ccValidate");
   
   if (getControlChars("certificateAuthlib", &ln)==0) {
-    libraryName(NULL,ln,dlName);
+    libraryName(NULL,ln,dlName, 512);
     if ((authLib=dlopen(dlName, RTLD_LAZY))) {
       validate= dlsym(authLib, "_sfcCertificateAuthenticate");
       if (validate) {
@@ -1624,7 +1899,7 @@ static int ccValidate(X509 *certificate, char ** principal, int mode)
     }
   } else {
     mlogf(M_ERROR,M_SHOW,
-	  "--- Certificate authentication exit not configured\n",dlName);	
+	  "--- Certificate authentication exit not configured\n");	
   }
   _SFCB_RETURN(result);
 }
