@@ -1,5 +1,5 @@
 /*
- * $Id: instance.c,v 1.40 2008/04/15 22:06:00 buccella Exp $
+ * $Id: instance.c,v 1.45 2008/11/21 20:23:51 mchasal Exp $
  *
  * © Copyright IBM Corp. 2005, 2007
  *
@@ -59,6 +59,7 @@ CMPIObjectPath *internal_new_CMPIObjectPath(int mode, const char *nameSpace,
                                             CMPIStatus * rc);
 extern const char *ClObjectGetClString(ClObjectHdr * hdr, ClString * id);
 extern int sfcb_comp_CMPIValue(CMPIValue *val1, CMPIValue *val2, CMPIType type);
+extern long addClString(ClObjectHdr * hdr, const char *str);
 
 extern CMPIBroker *Broker;
 
@@ -278,6 +279,57 @@ static CMPIStatus __ift_setProperty(const CMPIInstance * instance,
    CMReturn(CMPI_RC_OK);
 }
 
+static CMPIStatus __ift_setObjectPath(CMPIInstance * inst,
+                                      const CMPIObjectPath * cop)
+{
+   CMPIStatus tmp1, tmp2, tmp3;
+   CMPIString *str;
+   const char *ns, *cn;
+   int j;
+   CMPIStatus rc = { CMPI_RC_OK, NULL };
+
+   /* get information out of objectpath */
+   if (cop) {
+      j = CMGetKeyCount(cop, &tmp1);
+      str = CMGetClassName(cop, &tmp2);
+      cn = CMGetCharsPtr(str, NULL);
+      str = CMGetNameSpace(cop, &tmp3);
+      ns = CMGetCharsPtr(str, NULL);
+   }
+
+   /* there SHOULD be an op passed in, otherwise this call is useless.... */
+   else {
+      j=0;
+      //SFCB_ASM("int $3");
+      ns = "*NoNameSpace*";
+      cn = "*NoClassName*";
+      tmp1.rc=tmp2.rc=tmp3.rc=CMPI_RC_OK;
+   }
+
+   if (tmp1.rc != CMPI_RC_OK || tmp2.rc != CMPI_RC_OK || tmp3.rc != CMPI_RC_OK) {
+      rc.rc = CMPI_RC_ERR_FAILED;
+      return rc;
+   }
+   else {
+
+      /* set cn and ns in inst */
+      ClInstance* instance = (ClInstance*)inst->hdl;
+      if (ns) instance->nameSpace.id = addClString(&instance->hdr, ns);
+      if (cn) instance->className.id = addClString(&instance->hdr, cn);
+
+      /* loop over keys, set them in the inst */
+      while (j-- && (tmp1.rc == CMPI_RC_OK)) {
+         CMPIString *keyName;
+         CMPIData tmp = CMGetKeyAt(cop, j, &keyName, &tmp1);
+         __ift_setProperty(inst, CMGetCharsPtr(keyName, NULL),
+                           &tmp.value, tmp.type);
+      }
+      return tmp1;
+   }
+
+   return rc;
+
+}
 
 static CMPIObjectPath *__ift_getObjectPath(const CMPIInstance * instance,
                                            CMPIStatus * rc)
@@ -288,7 +340,6 @@ static CMPIObjectPath *__ift_getObjectPath(const CMPIInstance * instance,
    CMPIStatus tmp;
    const char *cn = ClInstanceGetClassName((ClInstance *) instance->hdl);
    const char *ns = ClInstanceGetNameSpace((ClInstance *) instance->hdl);
-   char *id;
 
    CMPIObjectPath *cop;
    cop = TrackedCMPIObjectPath(ns, cn, rc);
@@ -316,7 +367,8 @@ static CMPIObjectPath *__ift_getObjectPath(const CMPIInstance * instance,
       unsigned int e, m;
 
       if (mtx == NULL) {
-	 mtx = malloc(sizeof(CMPI_MUTEX_TYPE));
+         int dummy = 0;
+	 mtx = memAlloc(MEM_TRACKED, sizeof(CMPI_MUTEX_TYPE), &dummy);
 	 *mtx = Broker->xft->newMutex(0); 
       }
       Broker->xft->lockMutex(*mtx);
@@ -341,7 +393,6 @@ static CMPIObjectPath *__ift_getObjectPath(const CMPIInstance * instance,
 
       for (e = 0; e < m; e++) {
          CMPIString *n = kl->ft->getElementAt(kl, e, NULL).value.string;
-         id=CMGetCharPtr(n);
          d = __ift_getProperty(instance, CMGetCharPtr(n), &tmp);
          if (tmp.rc == CMPI_RC_OK) {
             CMAddKey(cop, CMGetCharPtr(n), &d.value, d.type);
@@ -384,7 +435,13 @@ static CMPIStatus __ift_setPropertyFilter(CMPIInstance * instance,
         for (j = 0, m = __ift_getPropertyCount(instance, &st); j < m; j++) {
 	    data = __ift_internal_getPropertyAt(instance, j, &name, &st, 1);
             if(__contained_list((char**)propertyList, name) || __contained_list((char**)keys, name)) {
-            newInstance->ft->setProperty(newInstance, name, &data.value, data.type);
+               if ( (data.state & ~CMPI_keyValue) != 0) {
+                  newInstance->ft->setProperty(
+                     newInstance, name, NULL, data.type);
+               } else {
+                  newInstance->ft->setProperty(
+                     newInstance, name, &data.value, data.type);
+               }
             }
         }
         
@@ -524,7 +581,8 @@ static CMPIInstanceFT ift = {
    __ift_getPropertyCount,
    __ift_setProperty,
    __ift_getObjectPath,
-   __ift_setPropertyFilter
+   __ift_setPropertyFilter,
+   __ift_setObjectPath
 };
 
 static struct {
@@ -661,6 +719,12 @@ CMPIInstance *internal_new_CMPIInstance(int mode, const CMPIObjectPath * cop,
    else {
       instance.instance.hdl = ClInstanceNew(ns, cn);
 
+#ifdef HAVE_DEFAULT_PROPERTIES
+   if(!override) {
+      instFillDefaultProperties(&instance,ns,cn);
+   }
+#endif
+
       while (j-- && (tmp1.rc == CMPI_RC_OK)) {
          CMPIString *keyName;
          CMPIData tmp = CMGetKeyAt(cop, j, &keyName, &tmp1);
@@ -673,12 +737,6 @@ CMPIInstance *internal_new_CMPIInstance(int mode, const CMPIObjectPath * cop,
    tInst=memAddEncObj(mode, &instance, sizeof(instance),&state);
    tInst->mem_state=state;
    tInst->refCount=0;
-   
-#ifdef HAVE_DEFAULT_PROPERTIES
-   if(!override) {
-      instFillDefaultProperties(tInst,ns,cn);
-   }
-#endif
 
    return (CMPIInstance*)tInst;
 }
