@@ -357,15 +357,6 @@ UtilList *SafeInternalProviderAddEnumInstances(UtilList *ul, CMPIInstanceMI * mi
    _SFCB_RETURN(ul);
 }
 
-CMPIStatus SafeInternalProviderEnumInstances(CMPIInstanceMI * mi, const CMPIContext * ctx, const CMPIResult * rslt,
-                                         const CMPIObjectPath * ref, const char **properties, int ignprov)
-{
-   CMPIStatus st;
-   _SFCB_ENTER(TRACE_INTERNALPROVIDER, "SafeInternalProviderEnumInstances");
-   st=enumInstances(mi,ctx,(void*)rslt,ref,properties,return2result,ignprov);
-   _SFCB_RETURN(st);
-}
-
 CMPIInstance *internalProviderGetInstance(const CMPIObjectPath * cop, CMPIStatus *rc)
 {
    int len;
@@ -469,12 +460,14 @@ CMPIStatus InternalProviderCreateInstance(CMPIInstanceMI * mi,
      if (isa(nss, cns, "cim_registeredprofile")) {
         CMPIArray* atArray;
         atArray = CMGetProperty(ci, "AdvertiseTypes", &st).value.array;
+#ifdef HAVE_SLP
         if (st.rc == CMPI_RC_OK || 
             atArray != NULL || 
             CMGetArrayElementAt(atArray, 0, &st).value.uint16 == 3) {
-
-          kill(slppid, SIGHUP);  /* restart SLP to update RegisteredProfiles */
+	  if (slppid > 1) /* sanity check */
+            kill(slppid, SIGHUP);  /* restart SLP to update RegisteredProfiles */
         }
+#endif
      }
    }
 
