@@ -56,6 +56,7 @@ static CMPIStatus okSt = { CMPI_RC_OK, NULL };
 static char *getSfcbUuid()
 {
    static char *uuid=NULL;
+   static char *u=NULL;
    
    if (uuid==NULL) {
       FILE *uuidFile;
@@ -76,14 +77,14 @@ static char *getSfcbUuid()
          }    
          fclose(uuidFile);
       }
-      else {
+      else if (u==NULL) {
          char hostName[512];
          gethostname(hostName,511);
-         char *u=(char*)malloc(strlen(hostName)+32);
+         u=(char*)malloc(strlen(hostName)+32);
          strcpy(u,"sfcb:NO-UUID-FILE-");
          strcat(u,hostName);
-         return u;
       }  
+      return u; 
    }   
    return uuid;
 }
@@ -220,7 +221,7 @@ static CMPIStatus NameSpaceProviderGetInstance(CMPIInstanceMI * mi,
       }
       else st.rc=CMPI_RC_ERR_NOT_FOUND;   
    }
-   else st.rc=CMPI_RC_ERR_NO_SUCH_PROPERTY;  
+   else st.rc=CMPI_RC_ERR_INVALID_PARAMETER;  
    
    _SFCB_RETURN(st);
 }
@@ -372,9 +373,7 @@ static CMPIStatus ObjectManagerProviderEnumInstances(CMPIInstanceMI * mi,
    str[0]=0;
    gethostname(str,511);
    CMSetProperty(ci,"SystemName",str,CMPI_chars);
-   char *name=getSfcbUuid();
    CMSetProperty(ci,"Name",getSfcbUuid(),CMPI_chars);
-   free(name);
    CMSetProperty(ci,"GatherStatisticalData",&bul,CMPI_boolean);
    CMSetProperty(ci,"ElementName","sfcb",CMPI_chars);
    CMSetProperty(ci,"Description",PACKAGE_STRING,CMPI_chars);
@@ -406,7 +405,7 @@ static CMPIStatus ObjectManagerProviderGetInstance(CMPIInstanceMI * mi,
          return ObjectManagerProviderEnumInstances(mi,ctx,rslt,ref,properties);
       else st.rc=CMPI_RC_ERR_NOT_FOUND;   
    }
-   else st.rc=CMPI_RC_ERR_NO_SUCH_PROPERTY;  
+   else st.rc=CMPI_RC_ERR_INVALID_PARAMETER;  
    
    _SFCB_RETURN(st);
 }
@@ -468,6 +467,8 @@ static CMPIStatus ComMechProviderEnumInstances(CMPIInstanceMI * mi,
    gethostname(hostName,511);
    CMSetProperty(ci,"SystemName",hostName,CMPI_chars);
    CMSetProperty(ci,"Name",getSfcbUuid(),CMPI_chars);
+   /* Version of CIM-XML that is supported */
+   CMSetProperty(ci,"Version","1.0",CMPI_chars);
    
    mech=2;
    CMSetProperty(ci,"CommunicationMechanism",&mech,CMPI_uint16);
@@ -505,7 +506,7 @@ static CMPIStatus ComMechProviderGetInstance(CMPIInstanceMI * mi,
          return ComMechProviderEnumInstances(mi,ctx,rslt,ref,properties);
       else st.rc=CMPI_RC_ERR_NOT_FOUND;   
    }
-   else st.rc=CMPI_RC_ERR_NO_SUCH_PROPERTY;  
+   else st.rc=CMPI_RC_ERR_INVALID_PARAMETER;  
    
    _SFCB_RETURN(st);
 }
