@@ -177,7 +177,7 @@ static void *ensureClSpace(ClObjectHdr * hdr, ClSection * sct, int size,
 }
 
 
-static long addClString(ClObjectHdr * hdr, const char *str)
+long addClString(ClObjectHdr * hdr, const char *str)
 {
    _SFCB_ENTER(TRACE_OBJECTIMPL, "addClString");
 
@@ -264,6 +264,8 @@ static long addClObject(ClObjectHdr * hdr, const void *obj, int size)
 	return(retCode);
 }
 
+
+/* TODO: search for duplicate code between here and addObjectPropertyH() */
 static long addClArray(ClObjectHdr * hdr, CMPIData d)
 {
    _SFCB_ENTER(TRACE_OBJECTIMPL, "addClArray");
@@ -609,6 +611,10 @@ static char *datatypeToString(CMPIData * d, char **array)
    case CMPI_chars:
    case CMPI_string:
       return "string";
+   case CMPI_real64:
+      return "real64";
+   case CMPI_real32:
+      return "real32";
    case CMPI_sint64:
       return "sint64";
    case CMPI_uint64:
@@ -627,26 +633,83 @@ static char *datatypeToString(CMPIData * d, char **array)
       return "sint8";
    case CMPI_boolean:
       return "boolean";
+   case CMPI_char16:
+      return "char16";
    case CMPI_ref:
-      return "reference";
+      return "ref";
+   case CMPI_instance:
+      return "instance";
    case CMPI_dateTime:
       return "dateTime";
+   default:
+      return "unknownType";
    }
-   return "*??*";
 }
 
 static char *dataValueToString(ClObjectHdr * hdr, CMPIData * d)
 {
    static char *True = "true", *False = "false";
+   // Allocate for the biggest numeric type we'll get, if 
+   // we get a char type, it's reallocated below.
+   char *retval=malloc(sizeof(d->value.real64));
    switch (d->type) {
-   case CMPI_chars:
-      return (char *) ClObjectGetClString(hdr, (ClString *) & d->value.chars);
-   case CMPI_sint32:
-      return "sint32";
-   case CMPI_boolean:
-      return d->value.boolean ? True : False;
+       case CMPI_chars:
+          {
+             char *ret = (char *) ClObjectGetClString(hdr, (ClString *) & d->value.chars);
+             if (ret) {
+                 retval=realloc(retval,sizeof(ret));
+                 strcpy(retval, ret); }
+             else
+                 return "\"\"";
+          }
+          break;
+       case CMPI_char16:
+          {
+             char *ret = (char *) ClObjectGetClString(hdr, (ClString *) & d->value.char16);
+             if (ret) {
+                 retval=realloc(retval,sizeof(ret));
+                 strcpy (retval, ret); }
+             else
+                 return "\"\"";
+          }
+          break;
+       case CMPI_real64:
+          sprintf(retval, "%lf", d->value.real64);
+          break;
+       case CMPI_real32:
+          sprintf(retval, "%f", d->value.real32);
+          break;
+       case CMPI_sint64:
+          sprintf(retval, "%lld", d->value.sint64);
+          break;
+       case CMPI_sint32:
+          sprintf(retval, "%d", d->value.sint32);
+          break;
+       case CMPI_sint16:
+          sprintf(retval, "%d", d->value.sint16);
+          break;
+       case CMPI_sint8:
+          sprintf(retval, "%d", d->value.sint8);
+          break;
+       case CMPI_uint64:
+          sprintf(retval, "%llu", d->value.uint64);
+          break;
+       case CMPI_uint32:
+          sprintf(retval, "%u", d->value.uint32);
+          break;
+       case CMPI_uint16:
+          sprintf(retval, "%u", d->value.uint16);
+          break;
+       case CMPI_uint8:
+          sprintf(retval, "%u", d->value.uint8);
+          break;
+       case CMPI_boolean:
+          strcpy(retval, d->value.boolean ? True : False);
+          break;
+       default:
+          strcpy(retval, "***??***");
    }
-   return "*??*";
+   return retval;
 }
 
 
@@ -782,8 +845,10 @@ static char *addQualifierToString(stringControl * sc, ClObjectHdr * hdr,
    cat2string(sc, ClObjectGetClString(hdr, &q->id));
    if (q->data.state != CMPI_nullValue) {
       cat2string(sc, "(");
-      cat2string(sc, dataValueToString(hdr, &q->data));
+      char * val=dataValueToString(hdr, &q->data);
+      cat2string(sc, val);
       cat2string(sc, ")");
+      free(val);
    }
    if (sb & 1) cat2string(sc, "]");
    return sc->str + o;
@@ -814,19 +879,21 @@ static int copyQualifiers(int ofs, int max, char *to, ClSection * ts,
 
 static int ClGetQualifierAt(ClClass * cls, ClQualifier *q, int id, CMPIData * data, char **name)
 {
-   if (data) *data = (q + id)->data;
-   if (data->type== CMPI_chars) {
+  if (data) { 
+    *data = (q + id)->data;
+    if (data->type== CMPI_chars) {
       const char *str =
-          ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
+        ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
       data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
       data->type = CMPI_string;
-   }
-   else if (data->type & CMPI_ARRAY) {
+    }
+    else if (data->type & CMPI_ARRAY) {
       data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&cls->hdr,
-            (ClArray *) & data->value.array);
-   }
-   if (name) *name = (char*)ClObjectGetClString(&cls->hdr, &(q + id)->id);
-   return 0;
+                                                            (ClArray *) & data->value.array);
+    }
+  }
+  if (name) *name = (char*)ClObjectGetClString(&cls->hdr, &(q + id)->id);
+  return 0;
 }
 
 int ClClassGetQualifierAt(ClClass * cls, int id, CMPIData * data, char **name)
@@ -1095,7 +1162,7 @@ int ClClassGetMethParamQualifierAt(ClClass * cls, ClParameter *parm, int id, CMP
    
    ClGetQualifierAt(cls,q,id,data,name); 
    
-   if (data->type & CMPI_ARRAY && data->value.array) {
+   if (data && data->type & CMPI_ARRAY && data->value.array) {
       data->value.array = native_make_CMPIArray((CMPIData *) data->value.array, 
          NULL, &cls->hdr);
    }   
@@ -1117,7 +1184,6 @@ int ClObjectLocateProperty(ClObjectHdr * hdr, ClSection * prps, const char *id)
    ClProperty *p;
 
    p=(ClProperty*)getSectionPtr(hdr,prps); 
-
    for (i = 0; i < prps->used; i++) {
       if (strcasecmp(id, ClObjectGetClString(hdr, &(p + i)->id)) == 0)
          return i + 1;
@@ -1148,6 +1214,9 @@ static char *addPropertyToString(stringControl * sc, ClObjectHdr * hdr,
    ClQualifier *q;
    char *array = NULL;
 
+   if (p->data.state && CMPI_nullValue)
+       return NULL;
+
    q = (ClQualifier *) ClObjectGetClSection(hdr, &p->qualifiers);
    if ((l = p->qualifiers.used)) {
       for (i = 0; i < l; sb = 0, i++) {
@@ -1158,14 +1227,22 @@ static char *addPropertyToString(stringControl * sc, ClObjectHdr * hdr,
       cat2string(sc, "\n");
    }
 
+
    cat2string(sc, " ");
    cat2string(sc, datatypeToString(&p->data, &array));
    cat2string(sc, " ");
    cat2string(sc, ClObjectGetClString(hdr, &p->id));
    if (array) cat2string(sc, array);
    cat2string(sc, " = ");
-   cat2string(sc, dataValueToString(hdr,&p->data));     
+
+   char *val = dataValueToString(hdr,&p->data);
+   if (val == NULL || *val == 0)
+       cat2string(sc, "\"\"");
+   else
+       cat2string(sc, val);
+
    cat2string(sc, ";\n");
+   free(val);
    return sc->str + o;
 }
 
@@ -1232,7 +1309,7 @@ static int addObjectPropertyH(ClObjectHdr * hdr, ClSection * prps,
    CMPIStatus st;
 
    _SFCB_ENTER(TRACE_OBJECTIMPL, "addObjectPropertyH");
-
+   /* if id is not in the section (prps), add it, increment number of props */
    if ((i = ClObjectLocateProperty(hdr, prps, id)) == 0) {
       p = (ClProperty *) ensureClSpace(hdr, prps, sizeof(*p), 8);
       p = p + (prps->used++);
@@ -1298,6 +1375,9 @@ static int addObjectPropertyH(ClObjectHdr * hdr, ClSection * prps,
             _SFCB_RETURN(-CMPI_RC_ERR_TYPE_MISMATCH);
             
          if (d.type != CMPI_chars) {
+            if (d.type != CMPI_string) {
+               _SFCB_RETURN(-CMPI_RC_ERR_TYPE_MISMATCH);
+	    }
             replaceClString(hdr, (long)od.value.chars, "");
             (p + i - 1)->data = d;
          }
@@ -1315,6 +1395,9 @@ static int addObjectPropertyH(ClObjectHdr * hdr, ClSection * prps,
       
       else if (od.type == CMPI_dateTime && (d.state & CMPI_nullValue) == 0) {
          char chars[26];
+         if (d.type != CMPI_dateTime) {
+           _SFCB_RETURN(-CMPI_RC_ERR_TYPE_MISMATCH);
+	 }
          dateTime2chars(d.value.dateTime, NULL, chars);
 	 if (od.value.chars) {
 	   (p + i - 1)->data = d;
@@ -1329,6 +1412,9 @@ static int addObjectPropertyH(ClObjectHdr * hdr, ClSection * prps,
 
       else if (od.type == CMPI_ref && (d.state & CMPI_nullValue) == 0) {
          char chars[4096] = { 0 };
+         if (d.type != CMPI_ref) {
+           _SFCB_RETURN(-CMPI_RC_ERR_TYPE_MISMATCH);
+	 }
          sfcb_pathToChars(d.value.ref, &st, chars);
 	 if (od.value.chars) {   
 	   (p + i - 1)->data = d;
@@ -1342,6 +1428,9 @@ static int addObjectPropertyH(ClObjectHdr * hdr, ClSection * prps,
       }
       
       else if ((od.type & CMPI_ARRAY) && (d.state & CMPI_nullValue) == 0) {
+        if (!(d.type & CMPI_ARRAY)) {
+           _SFCB_RETURN(-CMPI_RC_ERR_TYPE_MISMATCH);
+	}
 	if (od.value.array) {
 	  (p + i - 1)->data = d;
 	  replaceClArray(hdr, (long)od.value.array, d);
@@ -1589,7 +1678,38 @@ int ClClassGetPropertyAt(ClClass * cls, int id, CMPIData * data, char **name,
    p = (ClProperty *) ClObjectGetClSection(&cls->hdr, &cls->properties);
    if (id < 0 || id > cls->properties.used)
       return 1;
-   if (data) *data = (p + id)->data;
+    
+   if ((p + id)->quals & ClProperty_Q_EmbeddedObject) {
+     data->type = (data->type & CMPI_ARRAY ? CMPI_instance | CMPI_ARRAY : CMPI_instance); 
+   }
+   
+   if (data) { 
+     *data = (p + id)->data;
+     if (data->state & CMPI_nullValue)  {
+       data->value.uint64=0;
+     }
+     else if (data->type == CMPI_chars) {
+       const char *str =
+         ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
+       data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
+       data->type = CMPI_string;
+     }
+     else if (data->type == CMPI_dateTime) {
+       const char *str =
+         ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
+       data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL); 
+     }
+     else if (data->type & CMPI_ARRAY) {
+       data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&cls->hdr,
+                                             (ClArray *) & data->value.array);
+     }
+     else if (data->type == CMPI_instance) {
+       data->value.inst = 
+         relocateSerializedInstance((void *)ClObjectGetClObject(&cls->hdr,
+                                    (ClString *) & data->value.inst));          
+     }
+   }
+
    if (name) *name = (char*)ClObjectGetClString(&cls->hdr, &(p + id)->id);
    if (quals) *quals = (p + id)->quals;
    if (refName) {
@@ -1600,34 +1720,7 @@ int ClClassGetPropertyAt(ClClass * cls, int id, CMPIData * data, char **name,
          *refName = NULL;
       }
    }
-   
-   if ((p + id)->quals & ClProperty_Q_EmbeddedObject) {
-   	     data->type = (data->type & CMPI_ARRAY ? CMPI_instance | CMPI_ARRAY : CMPI_instance); 
-   }
-   
-   if (data->state & CMPI_nullValue)  {
-      data->value.uint64=0;
-   }
-   else if (data->type == CMPI_chars) {
-      const char *str =
-          ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
-      data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
-      data->type = CMPI_string;
-   }
-   else if (data->type == CMPI_dateTime) {
-      const char *str =
-         ClObjectGetClString(&cls->hdr, (ClString *) & data->value.chars);
-      data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL); 
-   }
-   else if (data->type & CMPI_ARRAY) {
-      data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&cls->hdr,
-            (ClArray *) & data->value.array);
-   }
-   else if (data->type == CMPI_instance) {
-      data->value.inst = 
-   	      relocateSerializedInstance((void *)ClObjectGetClObject(&cls->hdr,
-   	  							(ClString *) & data->value.inst));   	
-   }
+
    return 0;
 }
 
@@ -1757,6 +1850,14 @@ ClInstance *ClInstanceNew(const char *ns, const char *cn)
    return newInstanceH(ns, cn);
 }
 
+ClInstance *ClInstanceNewFromMof(const char *ns, const char *cn)
+{
+   ClInstance *inst = newInstanceH(ns, cn);
+   if (inst)
+       inst->hdr.flags |= HDR_FromMof;
+   return inst;
+}
+
 static long sizeInstanceH(ClObjectHdr * hdr, ClInstance * inst)
 {
    long sz = sizeof(*inst);
@@ -1842,7 +1943,7 @@ char *ClInstanceToString(ClInstance * inst)
 
    p = (ClProperty *) ClObjectGetClSection(&inst->hdr, &inst->properties);
    for (i = 0, l = inst->properties.used; i < l; i++) {
-      addPropertyToString(&sc, &inst->hdr, p + i);
+      addPropertyToString(&sc, &inst->hdr, (p+i));
    }
 
    cat2string(&sc, "};\n");
@@ -1865,31 +1966,35 @@ int ClInstanceGetPropertyAt(ClInstance * inst, int id, CMPIData * data,
    p = (ClProperty *) ClObjectGetClSection(&inst->hdr, &inst->properties);
    if (id < 0 || id > inst->properties.used)
       _SFCB_RETURN(1);
-   if (data) *data = (p + id)->data;
+   if (data) {
+     *data = (p + id)->data;
+     if (data->type == CMPI_chars) {
+       const char *str =
+         ClObjectGetClString(&inst->hdr, (ClString *) & data->value.chars);
+       data->value.string = sfcb_native_new_CMPIString(str, NULL, 2);
+       data->type = CMPI_string;
+     }
+     if (data->type == CMPI_dateTime) {
+       const char *str =
+         ClObjectGetClString(&inst->hdr, (ClString *) & data->value.chars);
+       data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL);
+     }
+     if (data->type & CMPI_ARRAY) {
+       data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&inst->hdr,
+                                                             (ClArray *) & data->value.array);
+     }
+     if (data->type == CMPI_instance) {
+       data->value.inst = (void *)ClObjectGetClObject(&inst->hdr,
+                                                      (ClString *) & data->value.inst);
+       if(data->value.inst) {
+         relocateSerializedInstance(data->value.inst);
+       }
+     }   
+   }
+
    if (name) *name = (char*)ClObjectGetClString(&inst->hdr, &(p + id)->id);
    if (quals) *quals = (p + id)->quals;
-   if (data->type == CMPI_chars) {
-      const char *str =
-          ClObjectGetClString(&inst->hdr, (ClString *) & data->value.chars);
-      data->value.string = sfcb_native_new_CMPIString(str, NULL, 2);
-      data->type = CMPI_string;
-   }
-   if (data->type == CMPI_dateTime) {
-      const char *str =
-          ClObjectGetClString(&inst->hdr, (ClString *) & data->value.chars);
-      data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL);
-   }
-   if (data->type & CMPI_ARRAY) {
-      data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&inst->hdr,
-            (ClArray *) & data->value.array);
-   }
-   if (data->type == CMPI_instance) {
-      data->value.inst = (void *)ClObjectGetClObject(&inst->hdr,
-      					(ClString *) & data->value.inst);
-      if(data->value.inst) {
-      	 relocateSerializedInstance(data->value.inst);
-      }
-   }   
+
    _SFCB_RETURN(0);
 }
 
@@ -2060,19 +2165,22 @@ int ClObjectPathGetKeyAt(ClObjectPath * op, int id, CMPIData * data,
 
    p = (ClProperty *) ClObjectGetClSection(&op->hdr, &op->properties);
    if (id < 0 || id > op->properties.used) return 1;
-   if (data) *data = (p + id)->data;
-   if (name) *name = (char*)ClObjectGetClString(&op->hdr, &(p + id)->id);
-   if (data->type == CMPI_chars) {
-      const char *str =
-          ClObjectGetClString(&op->hdr, (ClString *) & data->value.chars);
-      data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
-      data->type = CMPI_string;
-   }
-   else if (data->type == CMPI_dateTime) {
-      const char *str =
+   if (data) {
+     *data = (p + id)->data;
+     if (data->type == CMPI_chars) {
+       const char *str =
          ClObjectGetClString(&op->hdr, (ClString *) & data->value.chars);
-      data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL); 
+       data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
+       data->type = CMPI_string;
+     }
+     else if (data->type == CMPI_dateTime) {
+       const char *str =
+         ClObjectGetClString(&op->hdr, (ClString *) & data->value.chars);
+       data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL); 
+     }
    }
+   if (name) *name = (char*)ClObjectGetClString(&op->hdr, &(p + id)->id);
+
    return 0;
 }
 
@@ -2224,29 +2332,32 @@ int ClArgsGetArgAt(ClArgs * arg, int id, CMPIData * data, char **name)
    p = (ClProperty *) ClObjectGetClSection(&arg->hdr, &arg->properties);
    if (id < 0 || id > arg->properties.used)
       return 1;
-   if (data) *data = (p + id)->data;
+   if (data) {
+     *data = (p + id)->data;
+     if (data->type == CMPI_chars) {
+       const char *str = ClObjectGetClString(&arg->hdr, (ClString *) & data->value.chars);
+       data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
+       data->type = CMPI_string;
+     }
+     if (data->type == CMPI_dateTime) {
+       const char *str =
+         ClObjectGetClString(&arg->hdr, (ClString *) & data->value.chars);
+       data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL);
+     }
+     if (data->type & CMPI_ARRAY) {
+       data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&arg->hdr,
+                                             (ClArray *) & data->value.array);
+     }
+     if (data->type == CMPI_instance) {
+       data->value.inst = (void *)ClObjectGetClObject(&arg->hdr,
+                                             (ClString *) & data->value.inst);
+       if(data->value.inst) {
+         relocateSerializedInstance(data->value.inst);
+       }
+     }   
+   }
    if (name) *name = (char*)ClObjectGetClString(&arg->hdr, &(p + id)->id);
-   if (data->type == CMPI_chars) {
-      const char *str = ClObjectGetClString(&arg->hdr, (ClString *) & data->value.chars);
-      data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
-      data->type = CMPI_string;
-   }
-   if (data->type == CMPI_dateTime) {
-      const char *str =
-          ClObjectGetClString(&arg->hdr, (ClString *) & data->value.chars);
-      data->value.dateTime = sfcb_native_new_CMPIDateTime_fromChars(str, NULL);
-   }
-   if (data->type & CMPI_ARRAY) {
-      data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&arg->hdr,
-         (ClArray *) & data->value.array);
-   }
-   if (data->type == CMPI_instance) {
-   	  data->value.inst = (void *)ClObjectGetClObject(&arg->hdr,
-      								(ClString *) & data->value.inst);
-      if(data->value.inst) {
-      	 relocateSerializedInstance(data->value.inst);
-      }
-   }   
+
    _SFCB_RETURN(0);
 }
 
@@ -2351,17 +2462,19 @@ int ClQualifierAddQualifier(ClObjectHdr * hdr, ClSection * qlfs,
 
 static int ClGetQualifierFromQualifierDeclaration(ClQualifierDeclaration * q, ClQualifier *qData, CMPIData * data)
 {
-   if (data) *data = qData->data;
-   if (data->type == CMPI_chars) {
+  if (data) {
+    *data = qData->data;
+    if (data->type == CMPI_chars) {
       const char *str =
-          ClObjectGetClString(&q->hdr, (ClString *) & data->value.chars);
+        ClObjectGetClString(&q->hdr, (ClString *) & data->value.chars);
       data->value.string = sfcb_native_new_CMPIString(str, NULL, 0);
       data->type = CMPI_string;
-   }
-   else if (data->type & CMPI_ARRAY) {
+    }
+    else if (data->type & CMPI_ARRAY) {
       data->value.dataPtr.ptr = (void *) ClObjectGetClArray(&q->hdr,
-            (ClArray *) & data->value.array);
-   }
+                                                            (ClArray *) & data->value.array);
+    }
+  }
    return 0;
 }
 
