@@ -1,6 +1,6 @@
 
 /*
- * $Id: providerDrv.c,v 1.58 2008/04/09 18:43:14 smswehla Exp $
+ * $Id: providerDrv.c,v 1.77 2009/05/01 22:17:21 mchasal Exp $
  *
  * © Copyright IBM Corp. 2005, 2007
  *
@@ -129,10 +129,8 @@ extern CMPISelectExp *NewCMPISelectExp(const char *queryString, const char *lang
            const char *sns, CMPIArray ** projection, CMPIStatus * rc);
 NativeSelectExp *activFilters=NULL;
 extern void setStatus(CMPIStatus *st, CMPIrc rc, char *msg);
-extern CMPIString *NewCMPIString(const char *ptr, CMPIStatus * rc);
 
 static ProviderProcess *provProc=NULL,*curProvProc=NULL;
-static ProviderThread *curProvThread=NULL;
 static int provProcMax=0;
 static int idleThreadStartHandled=0;
 
@@ -158,17 +156,6 @@ typedef struct parms {
 } Parms;
 static Parms *activeThreadsFirst=NULL,*activeThreadsLast=NULL;
 
-struct providerThreadParams {
-  ProviderInfo* info;
-  pthread_t* thread_id;
-  //  ProviderProcess* proc;
-};
-
-pthread_mutex_t prov_thread_start_mutex;
-pthread_mutex_t prov_thread_start_mutex2;
-pthread_cond_t prov_thread_start_cv;
-pthread_cond_t prov_thread_start_cv2;
-
 /* old version support */
 typedef CMPIStatus (*authorizeFilterPreV1)
   (CMPIIndicationMI* mi, const CMPIContext* ctx, CMPIResult *result,
@@ -187,40 +174,40 @@ typedef CMPIStatus (*deActivateFilterPreV1)
    const CMPISelectExp* se, const  char *ns, const CMPIObjectPath* op, CMPIBoolean last);
 
 
-void libraryName(const char *dir, const char *location, char *fullName)
+void libraryName(const char *dir, const char *location, char *fullName, int buf_size)
 {
 #if defined(CMPI_PLATFORM_WIN32_IX86_MSVC)
   if (dir) {
-    sprintf(fullName,"%s\\%s.dll",dir,location);
+    snprintf(fullName, buf_size, "%s\\%s.dll",dir,location);
   } else {
-    sprintf(fullName,"%s.dll",location);
+    snprintf(fullName, buf_size, "%s.dll",location);
   }
 #elif defined(CMPI_PLATFORM_LINUX_GENERIC_GNU)
   if (dir) {
-   sprintf(fullName,"%s/lib%s.so",dir,location);
+    snprintf(fullName, buf_size, "%s/lib%s.so",dir,location);
   } else {
-    sprintf(fullName,"lib%s.so",location);
+    snprintf(fullName, buf_size, "lib%s.so",location);
   }
 #elif defined(CMPI_OS_HPUX)
   if (dir) {
-   sprintf(fullName,"%s/lib%s.so",dir,location);
+    snprintf(fullName, buf_size, "%s/lib%s.so",dir,location);
   } else {
-    sprintf(fullName,"lib%s.so",location);
+    snprintf(fullName, buf_size, "lib%s.so",location);
   }
 #elif defined(CMPI_OS_OS400)
   if (dir) {
-   strcpy(fullName, location);
+    strcpy(fullName, location);
 #elif defined(CMPI_OS_DARWIN)
   if (dir) {
-   sprintf(fullName,"%s/lib%s.dylib",dir,location);
+    snprintf(fullName, buf_size, "%s/lib%s.dylib",dir,location);
   } else {
-    sprintf(fullName,"lib%s.dylib",location);
+    snprintf(fullName, buf_size, "lib%s.dylib",location);
   }
 #else
   if (dir) {
-   sprintf(fullName,"%s/lib%s.so",dir,location);
+    snprintf(fullName, buf_size, "%s/lib%s.so",dir,location);
   } else {
-    sprintf(fullName,"lib%s.so",location);
+    snprintf(fullName, buf_size, "lib%s.so",location);
   }
 #endif
 }
@@ -312,7 +299,7 @@ static void handleSigSegv(int sig)
    mlogf(M_ERROR,M_SHOW, "-#- %s - %d provider exiting due to a SIGSEGV signal\n",
            processName, currentProc);
    while (threads) {
-      snprintf(msg,1023, "*** Provider %s(%d) exiting due to a SIGSEGV signal\n",
+      snprintf(msg,1023, "*** Provider %s(%d) exiting due to a SIGSEGV signal",
               threads->pInfo->providerName, currentProc);
       resp = errorCharsResp(CMPI_RC_ERR_FAILED, msg);
       sendResponse(threads->requestor, resp);
@@ -392,8 +379,8 @@ void* providerIdleThread()
          if (pInfo) {
             proc=curProvProc;
             if (proc) {
-               semAcquireUnDo(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
-               if ((val=semGetValue(sfcbSem,(proc->id*3)+provProcInuseId+provProcBaseId))==0) {            
+               semAcquireUnDo(sfcbSem,PROV_GUARD(proc->id));
+               if ((val=semGetValue(sfcbSem,PROV_INUSE(proc->id)))==0) {            
                   if ((now-proc->lastActivity)>provTimeoutInterval) {
                      ctx = native_new_CMPIContext(MEM_TRACKED,NULL);
                      noBreak=0;
@@ -428,7 +415,7 @@ void* providerIdleThread()
                      }
                   }
                }
-               semRelease(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
+               semRelease(sfcbSem,PROV_GUARD(proc->id));
             }      
          }
       }
@@ -441,94 +428,127 @@ void* providerIdleThread()
 }
 
 
-static int getInstanceMI(ProviderInfo *info, CMPIInstanceMI **mi, CMPIContext *ctx)
+static CMPIStatus getInstanceMI(ProviderInfo *info, CMPIInstanceMI **mi, CMPIContext *ctx)
 {
-    int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "getInstanceMI");
    
    if (info->instanceMI == NULL) info->instanceMI =
-          loadInstanceMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->instanceMI;
-   rc = info->instanceMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadInstanceMI(info->providerName, info->library, Broker, ctx, &st);
+   if (info->instanceMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->instanceMI;
+   }
+   _SFCB_RETURN(st);
 }
 
-static int getAssociationMI(ProviderInfo *info, CMPIAssociationMI **mi, CMPIContext *ctx)
+static CMPIStatus getAssociationMI(ProviderInfo *info, CMPIAssociationMI **mi, CMPIContext *ctx)
 {
-   int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "getAssociationMI");
     
    if (info->associationMI == NULL) info->associationMI =
-          loadAssociationMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->associationMI;
-   rc = info->associationMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadAssociationMI(info->providerName, info->library, Broker, ctx, 
+                            &st);
+   if (info->associationMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->associationMI;
+   }
+   _SFCB_RETURN(st);
 }
 
-static int getIndicationMI(ProviderInfo *info, CMPIIndicationMI **mi, CMPIContext *ctx)
+static CMPIStatus getIndicationMI(ProviderInfo *info, CMPIIndicationMI **mi, CMPIContext *ctx)
 {
-   int rc;
+    CMPIStatus st = { CMPI_RC_OK, NULL }; 
     _SFCB_ENTER(TRACE_PROVIDERDRV, "getIndicationMI");
     
    if (info->indicationMI == NULL) info->indicationMI =
-          loadIndicationMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->indicationMI;
-   rc = info->indicationMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadIndicationMI(info->providerName, info->library, Broker, ctx, &st);
+   if (info->indicationMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->indicationMI;
+   }
+   _SFCB_RETURN(st);
 }
 
-static int getMethodMI(ProviderInfo *info, CMPIMethodMI **mi, CMPIContext *ctx)
+static CMPIStatus getMethodMI(ProviderInfo *info, CMPIMethodMI **mi, CMPIContext *ctx)
 {
-   int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
     _SFCB_ENTER(TRACE_PROVIDERDRV, "getMethodMI");
     
    if (info->methodMI == NULL) info->methodMI =
-          loadMethodMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->methodMI;
-   rc = info->methodMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadMethodMI(info->providerName, info->library, Broker, ctx, &st);
+   if (info->methodMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->methodMI;
+   }
+   _SFCB_RETURN(st);
 }
 
-static int getClassMI(ProviderInfo *info, CMPIClassMI **mi, CMPIContext *ctx)
+static CMPIStatus getClassMI(ProviderInfo *info, CMPIClassMI **mi, CMPIContext *ctx)
 {
-   int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
     _SFCB_ENTER(TRACE_PROVIDERDRV, "getClassMI");
     
    if (info->classMI == NULL) info->classMI =
-          loadClassMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->classMI;
-   rc = info->classMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadClassMI(info->providerName, info->library, Broker, ctx, &st);
+   if (info->classMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->classMI;
+   }
+   _SFCB_RETURN(st);
 }
 
-static int getPropertyMI(ProviderInfo *info, CMPIPropertyMI **mi, CMPIContext *ctx)
+static CMPIStatus getPropertyMI(ProviderInfo *info, CMPIPropertyMI **mi, CMPIContext *ctx)
 {
-   int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
     _SFCB_ENTER(TRACE_PROVIDERDRV, "getPropertyMI");
     
    if (info->propertyMI == NULL) info->propertyMI =
-          loadPropertyMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->propertyMI;
-   rc = info->propertyMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadPropertyMI(info->providerName, info->library, Broker, ctx, &st);
+   if (info->propertyMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->propertyMI;
+   }
+   _SFCB_RETURN(st);
 }
 
 #ifdef HAVE_QUALREP
-static int getQualifierDeclMI(ProviderInfo *info, CMPIQualifierDeclMI **mi, CMPIContext *ctx)
+static CMPIStatus getQualifierDeclMI(ProviderInfo *info, CMPIQualifierDeclMI **mi, CMPIContext *ctx)
 {
-   int rc;
+   CMPIStatus st = { CMPI_RC_OK, NULL }; 
     _SFCB_ENTER(TRACE_PROVIDERDRV, "getQualiferMI");
     
    if (info->qualifierDeclMI == NULL) info->qualifierDeclMI =
-          loadQualifierDeclMI(info->providerName, info->library, Broker, ctx);
-   *mi = info->qualifierDeclMI;
-   rc = info->qualifierDeclMI ? 1 : 0;
-   _SFCB_RETURN(rc);
+          loadQualifierDeclMI(info->providerName, info->library, Broker, ctx, 
+                               &st);
+   if (info->qualifierDeclMI == NULL && st.rc == CMPI_RC_OK) {
+     st.rc = CMPI_RC_ERR_FAILED; 
+   }
+   else {
+     *mi = info->qualifierDeclMI;
+   }
+   _SFCB_RETURN(st);
 }
 #endif
 
 
-
+/* see if we're under the max allowed number of loaded providers.
+   if so, fork() for the new provider
+   I think this can be broken down into smaller functions
+*/
 static int getProcess(ProviderInfo * info, ProviderProcess ** proc)
 {
    int i;
@@ -545,9 +565,9 @@ static int getProcess(ProviderInfo * info, ProviderProcess ** proc)
      for (i = 0; i < provProcMax; i++) {
        if ((provProc+i) && provProc[i].pid &&
            provProc[i].group && strcmp(provProc[i].group,info->group)==0) {
-         semAcquire(sfcbSem,(provProc[i].id*3)+provProcGuardId+provProcBaseId);
-         semRelease(sfcbSem,(provProc[i].id*3)+provProcInuseId+provProcBaseId);
-         semRelease(sfcbSem,(provProc[i].id*3)+provProcGuardId+provProcBaseId);
+         semAcquire(sfcbSem,PROV_GUARD(provProc[i].id));
+         semRelease(sfcbSem,PROV_INUSE(provProc[i].id));
+         semRelease(sfcbSem,PROV_GUARD(provProc[i].id));
          info->pid=provProc[i].pid;
          info->providerSockets=provProc[i].providerSockets;
          _SFCB_TRACE(1,("--- Process %d shared by %s and %s",provProc[i].pid,info->providerName,
@@ -602,12 +622,12 @@ static int getProcess(ProviderInfo * info, ProviderProcess ** proc)
             info->proc=*proc;
             info->pid=currentProc; 
                          
-            semSetValue(sfcbSem,((*proc)->id*3)+provProcGuardId+provProcBaseId,0);
-            semSetValue(sfcbSem,((*proc)->id*3)+provProcInuseId+provProcBaseId,0);
-            semSetValue(sfcbSem,((*proc)->id*3)+provProcAliveId+provProcBaseId,0);
-            semReleaseUnDo(sfcbSem,((*proc)->id*3)+provProcAliveId+provProcBaseId);
-            semReleaseUnDo(sfcbSem,((*proc)->id*3)+provProcInuseId+provProcBaseId);
-            semRelease(sfcbSem,((*proc)->id*3)+provProcGuardId+provProcBaseId);
+            semSetValue(sfcbSem,PROV_GUARD((*proc)->id),0);
+            semSetValue(sfcbSem,PROV_INUSE((*proc)->id),0);
+            semSetValue(sfcbSem,PROV_ALIVE((*proc)->id),0);
+            semReleaseUnDo(sfcbSem,PROV_ALIVE((*proc)->id));
+            semReleaseUnDo(sfcbSem,PROV_INUSE((*proc)->id));
+            semRelease(sfcbSem,PROV_GUARD((*proc)->id));
             
             processProviderInvocationRequests(info->providerName);
             _SFCB_RETURN(0);
@@ -630,9 +650,6 @@ static int getProcess(ProviderInfo * info, ProviderProcess ** proc)
 // something like isLoaded() and doForkProvider()
 int forkProvider(ProviderInfo * info, OperationHdr * req, char **msg)
 {
-#ifdef USE_THREADS
-   return createProviderThread(info, req, msg);
-#endif
    _SFCB_ENTER(TRACE_PROVIDERDRV, "forkProvider");
    ProviderProcess *proc;
    ProviderInfo * pInfo;
@@ -640,15 +657,15 @@ int forkProvider(ProviderInfo * info, OperationHdr * req, char **msg)
 
    if (info->pid ) {
       proc=info->proc;
-      semAcquire(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
-      if ((val=semGetValue(sfcbSem,(proc->id*3)+provProcAliveId+provProcBaseId))) {
-         semRelease(sfcbSem,(proc->id*3)+provProcInuseId+provProcBaseId);
-         semRelease(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
+      semAcquire(sfcbSem,PROV_GUARD(proc->id));
+      if ((val=semGetValue(sfcbSem,PROV_ALIVE(proc->id)))) {
+         semRelease(sfcbSem,PROV_INUSE(proc->id));
+         semRelease(sfcbSem,PROV_GUARD(proc->id));
          _SFCB_TRACE(1, ("--- Provider %s still loaded",info->providerName));
          _SFCB_RETURN(CMPI_RC_OK)
       }
 
-      semRelease(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
+      semRelease(sfcbSem,PROV_GUARD(proc->id));
       _SFCB_TRACE(1, ("--- Provider has been unloaded prevously, will reload"));
       
       info->pid=0;
@@ -688,264 +705,22 @@ int forkProvider(ProviderInfo * info, OperationHdr * req, char **msg)
 
       resp = invokeProvider(&binCtx);
       resp->rc--;
-      if (resp->rc) {
-         *msg = strdup((char *) resp->object[0].data);
-      }
+      if (msg) {
+        if (resp->rc) {
+          *msg = strdup((char *) resp->object[0].data);
+        }
       else *msg = NULL;
+      }
 
       rc=resp->rc;
       _SFCB_TRACE(1, ("--- rc: %d", resp->rc));
       
-      free(resp);
+      if(resp) free(resp);
       _SFCB_RETURN(rc);
    }
    _SFCB_RETURN(CMPI_RC_ERR_FAILED);
 }
 
-void* startProviderThread(void* params) {
-
-  pthread_mutex_init(&prov_thread_start_mutex, NULL);
-  pthread_cond_init (&prov_thread_start_cv, NULL);
-
-  struct providerThreadParams* p = (struct providerThreadParams*)params;
-  ProviderInfo* info = p->info;
-  *(p->thread_id) = pthread_self();
-
-  pthread_mutex_lock(&prov_thread_start_mutex);
-  pthread_cond_signal(&prov_thread_start_cv);
-  pthread_mutex_unlock(&prov_thread_start_mutex);
-
-  //now wait for getThread to finish
-  pthread_mutex_lock(&prov_thread_start_mutex);
-  pthread_cond_wait(&prov_thread_start_cv2, &prov_thread_start_mutex);
-  pthread_mutex_unlock(&prov_thread_start_mutex);
-
-  processProviderInvocationRequests(info->providerName);
-
-  pthread_exit(NULL);
-}
-
-
-// if getProcess() can be broken into smaller functions, some could
-// be shared with getThread()
-static int getThread(ProviderInfo * info, ProviderThread ** thread) {
-
-   int i, pcrc;
-   static int seq=0;
-
-   _SFCB_ENTER(TRACE_PROVIDERDRV, "getThread");
-
-   if (provAutoGroup && info->group == NULL) {
-     /* implicitly put all providers in a module in a virtual group */
-     info->group = strdup(info->location);
-   }
-   
-   if (info->group) {
-     /* are we under the amount of maximum processes allowed? */
-     for (i = 0; i < provProcMax; i++) {
-       /* what is all this? if we use threads, do we even need process sharing? */
-       /* if ((provThread+i) && provThread[i].tid &&
-           provThread[i].group && strcmp(provThread[i].group,info->group)==0) {
-         semAcquire(sfcbSem,(provThread[i].id*3)+provProcGuardId+provProcBaseId);
-         semRelease(sfcbSem,(provThread[i].id*3)+provProcInuseId+provProcBaseId);
-         semRelease(sfcbSem,(provThread[i].id*3)+provProcGuardId+provProcBaseId);
-         info->tid=provThread[i].tid;
-         info->providerSockets=provThread[i].providerSockets;
-         _SFCB_TRACE(1,("--- Process %d shared by %s and %s",provThread[i].tid,info->providerName,
-			provThread[i].firstProv->providerName));
-         if (provThread[i].firstProv) info->next=provThread[i].firstProv;
-         else info->next = NULL;
-         provThread[i].firstProv=info;
-         info->thread=provThread+i;
-         if (info->unload<provThread[i].unload) provThread[i].unload=info->unload;
-         _SFCB_RETURN(provThread[i].tid);
-	 } */
-     }
-   }
-   for (i = 0; i < provProcMax; i++) {
-
-     // does provider #i already exist?
-     //if (pthread_equal(provThread[i].tid, 0)) {
-
-     /*
-
-     // fork() would have happened here
-
-     // all providers are in the saname thread, no need to keep calling setSignal
-         currentProc=getpid();
-         setSignal(SIGCHLD, SIG_DFL,0);
-         setSignal(SIGTERM, SIG_IGN,0);
-         setSignal(SIGHUP, SIG_IGN,0);
-         setSignal(SIGUSR1, handleSigUsr1,0);
-            
-         setSignal(SIGSEGV, handleSigSegv,SA_ONESHOT);
-     */
-            
-
-         _SFCB_TRACE(1,("--- Forked started for %s %d %d-%lu",
-                       info->providerName, curProvThread,providerSockets.receive,
-                       getInode(providerSockets.receive)));
-         processName=info->providerName;
-	 //does it matter if this is set?
-         providerProcess=1;
-         info->thread=*thread;
-         info->pid=currentProc; 
-         //the tid should probably be set as well...
-            
-         //again, is this necessary when using threads?
-         semSetValue(sfcbSem,((*thread)->id*3)+provProcGuardId+provProcBaseId,0);
-         semSetValue(sfcbSem,((*thread)->id*3)+provProcInuseId+provProcBaseId,0);
-         semSetValue(sfcbSem,((*thread)->id*3)+provProcAliveId+provProcBaseId,0);
-         semReleaseUnDo(sfcbSem,((*thread)->id*3)+provProcAliveId+provProcBaseId);
-         semReleaseUnDo(sfcbSem,((*thread)->id*3)+provProcInuseId+provProcBaseId);
-         semRelease(sfcbSem,((*thread)->id*3)+provProcGuardId+provProcBaseId);
-            
-         /* needed for thread wait */
-         pthread_mutex_init(&prov_thread_start_mutex, NULL);
-         pthread_cond_init (&prov_thread_start_cv, NULL);
-
-         pthread_t providerThread;
-	 //this could be a problem, as pthread_t != int on all systems
-	 pthread_t* thread_id = NULL;
-
-         struct providerThreadParams ptparams = {info, thread_id};
-
-         pcrc = pthread_create(&providerThread, NULL, &startProviderThread, &ptparams);
-
-	 //wait for thread to get going
-	 pthread_mutex_trylock(&prov_thread_start_mutex);
-	 pthread_cond_wait(&prov_thread_start_cv, &prov_thread_start_mutex);
-	 pthread_mutex_unlock(&prov_thread_start_mutex);
-
-         (*thread)->tid = info->tid = providerThread;
-     	 (*thread)->tid = info->tid = *thread_id;
-
-	 providerSockets=sPairs[(*thread)->id];
-
-	 (*thread)->providerSockets = info->providerSockets = providerSockets;
-	 (*thread)->group = info->group;
-	 (*thread)->unload=info->unload;
-	 (*thread)->firstProv=info;
-	 info->thread = *thread;
-	 info->next = NULL;
-         curProvThread=(*thread);
-         resultSockets=sPairs[(*thread)->id+ptBase];
-
-	 pthread_cond_signal(&prov_thread_start_cv2);
-
-         // from else block of if (info->pid == 0)
-         info->startSeq=++seq;
-
-	 _SFCB_TRACE(1,("--- Create thread for provider OK %s %d %d", info->providerName,
-		         info->tid, i));
-
-	  //should there be a break somewhere around here?
-       /*
-       }
-
-       else {
-         info->startSeq=++seq;
-       } 
-       */
-         _SFCB_TRACE(1,("--- Fork provider OK %s %d %d", info->providerName,
-                      info->pid, i));
-
-	 pthread_mutex_destroy(&prov_thread_start_mutex);
-	 pthread_cond_destroy(&prov_thread_start_cv);
-
-	 //should we be returning tid?  pcrc?
-         _SFCB_RETURN(info->pid);
-      }
-
-   *thread = NULL;
-   _SFCB_RETURN(-1);
-}
-
-int createProviderThread(ProviderInfo * info, OperationHdr * req, char **msg) {
-
-   _SFCB_ENTER(TRACE_PROVIDERDRV, "createProviderThread");
-   ProviderThread *thread = NULL;
-   ProviderInfo * pInfo;
-   //int val;
-   int rc;
-
-   if (info->pid ) {
-      thread=info->thread;
-   /* fror now, let's assume it needs to be reloaded
-      semAcquire(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
-      if ((val=semGetValue(sfcbSem,(proc->id*3)+provProcAliveId+provProcBaseId))) {
-         semRelease(sfcbSem,(proc->id*3)+provProcInuseId+provProcBaseId);
-         semRelease(sfcbSem,(proc->id*3)+provProcGuardId+provProcBaseId);
-         _SFCB_TRACE(1, ("--- Provider %s still loaded",info->providerName));
-         _SFCB_RETURN(CMPI_RC_OK)
-      }
-   */
-
-      semRelease(sfcbSem,(thread->id*3)+provProcGuardId+provProcBaseId);
-      _SFCB_TRACE(1, ("--- Provider has been unloaded prevously, will reload"));
-      
-      info->tid=0;
-      for (pInfo=thread->firstProv; pInfo; pInfo=pInfo->next) {
-         pInfo->pid=0;
-      }      
-      thread->firstProv=NULL;            
-      thread->tid=0; 
-      thread->group=NULL; 
-   }
-
-   _SFCB_TRACE(1, ("--- Creating thread for provider %s", info->providerName));
-
-   if (getThread(info, &thread) > 0) {
-
-      LoadProviderReq sreq = BINREQ(OPS_LoadProvider, 3);
-
-      BinRequestContext binCtx;
-      BinResponseHdr *resp;
-
-      memset(&binCtx,0,sizeof(BinRequestContext));
-      sreq.className = setCharsMsgSegment(info->className);
-      sreq.libName = setCharsMsgSegment(info->location);
-      sreq.provName = setCharsMsgSegment(info->providerName);
-      sreq.hdr.flags = info->type;
-      sreq.unload = info->unload;
-      sreq.hdr.provId = getProvIds(info).ids;
-
-      if (req) binCtx.oHdr = (OperationHdr *) req;
-      binCtx.bHdr = &sreq.hdr;
-      binCtx.bHdrSize = sizeof(sreq);
-      binCtx.provA.socket = info->providerSockets.send;
-      binCtx.provA.ids = getProvIds(info);
-      binCtx.chunkedMode=binCtx.xmlAs=binCtx.noResp=0;
-
-      _SFCB_TRACE(1, ("--- Invoke loader"));
-
-      resp = invokeProvider(&binCtx);
-      resp->rc--;
-      if (resp->rc) {
-         *msg = strdup((char *) resp->object[0].data);
-      }
-      else *msg = NULL;
-
-      rc=resp->rc;
-      _SFCB_TRACE(1, ("--- rc: %d", resp->rc));
-      
-      free(resp);
-      _SFCB_RETURN(rc);
-   }
-   _SFCB_RETURN(CMPI_RC_ERR_FAILED);
-}
-
-
-/* -------------
- * ---
- *      Provider Driver
- * ---
- * -------------
- */
-
-
-
-
 typedef struct provHandler {
    BinResponseHdr *(*handler) (BinRequestHdr *, ProviderInfo * info, int requestor);
 } ProvHandler;
@@ -1052,7 +827,6 @@ static int sendResponse(int requestor, BinResponseHdr * hdr)
 #endif
       default:
          mlogf(M_ERROR,M_SHOW,"--- bad sendResponse request %d\n", hdr->object[i].type);
-         *((char *) (void *) 0) = 0;
          _SFCB_ABORT();
       }
    }
@@ -1069,6 +843,7 @@ static int sendResponse(int requestor, BinResponseHdr * hdr)
 int sendResponseChunk(CMPIArray *r,int requestor, CMPIType type)
 {
    int i, count;
+   int rslt;
    BinResponseHdr *resp;
       
    _SFCB_ENTER(TRACE_PROVIDERDRV, "sendResponseChunk");
@@ -1086,7 +861,9 @@ int sendResponseChunk(CMPIArray *r,int requestor, CMPIType type)
       else resp->object[i] =
           setObjectPathMsgSegment(CMGetArrayElementAt(r, i, NULL).value.ref);
 
-   _SFCB_RETURN(sendResponse(requestor, resp));
+   rslt = sendResponse(requestor, resp);
+   if(resp) free(resp);
+   _SFCB_RETURN(rslt);
 }
 
 static BinResponseHdr *errorResp(CMPIStatus * rci)
@@ -1513,6 +1290,10 @@ static BinResponseHdr *getProperty(BinRequestHdr * hdr, ProviderInfo * info,
    ctx->ft->addEntry(ctx,CMPIPrincipal,(CMPIValue*)req->principal.data,CMPI_chars);
    ctx->ft->addEntry(ctx,"CMPISessionId",(CMPIValue*)&hdr->sessionId,CMPI_uint32);
 
+   if (info->propertyMI == NULL) info->propertyMI =
+                 loadPropertyMI(info->providerName, info->library, Broker, ctx,
+                                &rci); 
+
    _SFCB_TRACE(1, ("--- Calling provider %s",info->providerName));
    TIMING_START(hdr,info)
    rci = info->propertyMI->ft->getProperty(info->propertyMI, ctx, result, path, (const char*)req->name.data);
@@ -1560,6 +1341,10 @@ static BinResponseHdr *setProperty(BinRequestHdr * hdr, ProviderInfo * info,
    
    data = inst->ft->getPropertyAt(inst, 0, &pName, NULL);
 
+   if (info->propertyMI == NULL) info->propertyMI =
+                 loadPropertyMI(info->providerName, info->library, Broker, ctx,
+                                &rci);
+
    _SFCB_TRACE(1, ("--- Calling provider %s",info->providerName));
    TIMING_START(hdr,info)
    rci = info->propertyMI->ft->setProperty(info->propertyMI, ctx, result, path, (const char*)pName->hdl, data);
@@ -1680,15 +1465,22 @@ static BinResponseHdr *getInstance(BinRequestHdr * hdr, ProviderInfo * info,
    r = native_result2array(result);
 
    if (rci.rc == CMPI_RC_OK) {
-      count = 1;
-      resp = (BinResponseHdr *) calloc(1,sizeof(BinResponseHdr) +
-                                    ((count - 1) * sizeof(MsgSegment)));
-      resp->moreChunks=0;
-      resp->rc = 1;
-      resp->count = count;
-      for (i = 0; i < count; i++)
-         resp->object[i] =
-             setInstanceMsgSegment(CMGetArrayElementAt(r, i, NULL).value.inst);
+      if (r && CMGetArrayCount(r, NULL) > 0) {
+         count = 1;
+
+         resp = (BinResponseHdr *) calloc(1,sizeof(BinResponseHdr) +
+                                       ((count - 1) * sizeof(MsgSegment)));
+         resp->moreChunks=0;
+         resp->rc = 1;
+         resp->count = count;
+         for (i = 0; i < count; i++)
+            resp->object[i] =
+              setInstanceMsgSegment(CMGetArrayElementAt(r, i, NULL).value.inst);
+      } else {
+         rci.rc = CMPI_RC_ERR_NOT_FOUND;
+         rci.msg = NULL;
+         resp = errorResp(&rci);
+      }
    }
    else resp = errorResp(&rci);
    if (props) free(props);
@@ -1931,7 +1723,7 @@ CMPIValue queryGetValue(QLPropertySource* src, char* name, QLOpd *type)
    CMPIValue v={(long long)0};
    
    if (rc.rc==CMPI_RC_OK) {
-      if (d.type & CMPI_SINT) { 
+      if ((d.type & CMPI_SINT) == CMPI_SINT) { 
          if (d.type==CMPI_sint32) v.sint64=d.value.sint32;
          else if (d.type==CMPI_sint16) v.sint64=d.value.sint16;
          else if (d.type==CMPI_sint8) v.sint64=d.value.sint8;
@@ -2364,7 +2156,7 @@ static BinResponseHdr *deactivateFilter(BinRequestHdr *hdr, ProviderInfo* info,
    resp->rc=1;
 
    _SFCB_TRACE(1, ("---  pid: %d activFilters %p",currentProc,activFilters));
-   if (activFilters==NULL) _SFCB_RETURN(resp); 
+   if (info->indicationMI==NULL || activFilters==NULL) _SFCB_RETURN(resp); 
 
    for (se = activFilters; se; se = se->next) {
       if (se->filterId == req->filterId) {
@@ -2506,11 +2298,24 @@ static BinResponseHdr *opNotSupported(BinRequestHdr * hdr, ProviderInfo * info,
 }
 
 
+void appendStr(char **base, const char* str, const char* arg)
+{
+    if (*base == NULL) {
+        *base = sfcb_snprintf("%s %s", str, arg); 
+    }
+    else {
+        char* tmp = sfcb_snprintf("%s; %s %s", *base, str, arg); 
+        free(*base); 
+        *base = tmp; 
+    }
+}
 
-int initProvider(ProviderInfo *info, unsigned int sessionId)
+int initProvider(ProviderInfo *info, unsigned int sessionId, char** errorStr)
 {
    CMPIInstanceMI *mi = NULL;
    int rc=0;
+   CMPIStatus st; 
+   char* errstr = NULL; 
    unsigned int flgs=0;
    CMPIContext *ctx = native_new_CMPIContext(MEM_TRACKED,info);
    
@@ -2528,45 +2333,92 @@ int initProvider(ProviderInfo *info, unsigned int sessionId)
      
      
      if (info->type & INSTANCE_PROVIDER) {
-       rc |= (getInstanceMI(info, &mi, ctx) != 1);
+         st = getInstanceMI(info, &mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+            appendStr(&errstr, "Error from Instance MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }   
      if (info->type & ASSOCIATION_PROVIDER) {
-       rc |= (getAssociationMI(info, (CMPIAssociationMI **) & mi, ctx) != 1);
+         st = getAssociationMI(info, (CMPIAssociationMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Association MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }   
      if (info->type & METHOD_PROVIDER) {
-       rc |= (getMethodMI(info, (CMPIMethodMI **) & mi, ctx) != 1);
+         st = getMethodMI(info, (CMPIMethodMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Method MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }   
      if (info->type & INDICATION_PROVIDER) {
-       rc |= (getIndicationMI(info, (CMPIIndicationMI **) & mi, ctx) != 1);
+         st = getIndicationMI(info, (CMPIIndicationMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Indication MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }   
      if (info->type & CLASS_PROVIDER) {
-       rc |= (getClassMI(info, (CMPIClassMI **) & mi, ctx) != 1);
+         st = getClassMI(info, (CMPIClassMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Class MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }
      if (info->type & PROPERTY_PROVIDER) {
-       rc |= (getPropertyMI(info, (CMPIPropertyMI **) & mi, ctx) != 1);
+         st = getPropertyMI(info, (CMPIPropertyMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Property MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }
 #ifdef HAVE_QUALREP   
      if (info->type & QUALIFIER_PROVIDER) {
-       rc |= (getQualifierDeclMI(info, (CMPIQualifierDeclMI **) & mi, ctx) != 1);
+         st = getQualifierDeclMI(info, (CMPIQualifierDeclMI **) & mi, ctx); 
+         rc |= st.rc; 
+         if (st.rc != CMPI_RC_OK && st.msg != NULL) {
+             appendStr(&errstr, "Error from Qualifier MI Factory:", 
+                      CMGetCharsPtr(st.msg,NULL)); 
+         }
      }
 #endif   
      
      if (rc) {
        rc = -2;
+       if (errstr != NULL) {
+          *errorStr = sfcb_snprintf("Error initializing provider %s from %s for class %s.  %s", 
+                           info->providerName, info->location, 
+                           info->className, errstr);
+       }
+       else {
+          *errorStr = sfcb_snprintf("Error initializing provider %s from %s for class %s.", 
+                           info->providerName, info->location, info->className);
+       }
      } else {
        info -> initialized = 1;
+       *errorStr = NULL;
      }
      pthread_mutex_unlock(&info->initMtx);
    }
+   if (errstr != NULL) free(errstr); 
 
    _SFCB_RETURN(rc);
 }
 
 
-static int doLoadProvider(ProviderInfo *info, char *dlName)
+static int doLoadProvider(ProviderInfo *info, char *dlName, int dlName_length)
 {
    char *dirs,*dir,*dirlast, *dircpy;
    char *fullname;
+   int fullname_max_length = 0;
    struct stat stbuf;
 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "doLoadProvider");
@@ -2577,16 +2429,17 @@ static int doLoadProvider(ProviderInfo *info, char *dlName)
    }
    
 
-   libraryName(NULL, (char *) info->location, dlName);
+   libraryName(NULL, (char *) info->location, dlName, dlName_length);
    
    dircpy = strdup(dirs);
-   fullname = malloc(strlen(dircpy)+strlen(dlName)+2); /* sufficient */
+   fullname_max_length = strlen(dircpy)+strlen(dlName)+2; /* sufficient */
+   fullname = malloc(fullname_max_length);
    dir=strtok_r(dircpy," \t",&dirlast);
    info->library = NULL;
    while (dir) {
-     libraryName(dir, (char *) info->location, fullname);
+     libraryName(dir, (char *) info->location, fullname, fullname_max_length);
      if (stat(fullname,&stbuf) == 0) {
-       info->library = dlopen(fullname, RTLD_NOW);
+       info->library = dlopen(fullname, RTLD_NOW | RTLD_GLOBAL);
        if (info->library == NULL) {
 	 mlogf(M_ERROR,M_SHOW,"*** dlopen error: %s\n",dlerror());
        } else {
@@ -2621,7 +2474,7 @@ static BinResponseHdr *loadProvider(BinRequestHdr * hdr, ProviderInfo * info, in
    BinResponseHdr *resp;
    char dlName[512];
 
-   _SFCB_TRACE(1, ("--- Loading Provide %s %s %s", (char *) req->className.data,
+   _SFCB_TRACE(1, ("--- Loading provider %s %s %s", (char *) req->className.data,
                 (char *) req->provName.data, (char *) req->libName.data));
 
    info = (ProviderInfo *) calloc(1, sizeof(*info));
@@ -2634,21 +2487,21 @@ static BinResponseHdr *loadProvider(BinRequestHdr * hdr, ProviderInfo * info, in
    info->providerSockets = providerSockets;
    info->provIds.ids = hdr->provId;
 
-   switch (doLoadProvider(info,dlName)) {
+   switch (doLoadProvider(info,dlName, 512)) {
    case -1: {
       char msg[740];
-      sprintf(msg, "*** Failed to load %s for %s\n", dlName,
+      snprintf(msg,739,"*** Failed to load %s for %s", dlName,
               info->providerName);
-      mlogf(M_ERROR,M_SHOW,msg);
+      mlogf(M_ERROR,M_SHOW,"%s\n",msg);
       resp = errorCharsResp(CMPI_RC_ERR_FAILED, msg);
       free(info);
       _SFCB_RETURN(resp);
    }
    case -2:  {
       char msg[740];
-      sprintf(msg, "*** Inconsistent provider registration for %s (1)\n",
+      snprintf(msg,739,"*** Inconsistent provider registration for %s (1)",
               info->providerName);
-      mlogf(M_ERROR,M_SHOW,msg);
+      mlogf(M_ERROR,M_SHOW,"%s\n",msg);
       resp = errorCharsResp(CMPI_RC_ERR_FAILED, msg);
       free(info);
       _SFCB_RETURN(resp);
@@ -2760,9 +2613,10 @@ static void *processProviderInvocationRequestsThread(void *prms)
    Parms *parms = (Parms *) prms;
    BinRequestHdr *req = parms->req;
    int i,requestor=0,initRc=0;
+   char* errstr = NULL; 
 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "processProviderInvocationRequestsThread");
- 
+
    for (i = 0; i < req->count; i++)
       if (req->object[i].length)
          req->object[i].data=(void*)((long)req->object[i].data+(char*)req);
@@ -2790,11 +2644,11 @@ static void *processProviderInvocationRequestsThread(void *prms)
       if (pInfo && pInfo->library==NULL) { 
          char dlName[512];
          mlogf(M_INFO,M_SHOW,"--- Reloading provider\n");
-         doLoadProvider(pInfo,dlName);
+         doLoadProvider(pInfo,dlName, 512);
       }  
       
       if (pInfo && pInfo->initialized == 0) {
-	initRc=initProvider(pInfo,req->sessionId);
+	initRc=initProvider(pInfo,req->sessionId, &errstr);
 	_SFCB_TRACE(1, ("--- Provider initialization rc %d",initRc));
       }
 
@@ -2802,16 +2656,15 @@ static void *processProviderInvocationRequestsThread(void *prms)
    else pInfo = NULL;
 
    if (initRc) {
-      char msg[1024];
-      snprintf(msg,1023, "*** Inconsistent provider registration for %s (2)\n",
-              pInfo->providerName);
-      mlogf(M_ERROR,M_SHOW,msg);
-      _SFCB_TRACE(1, (msg));
-      resp = errorCharsResp(CMPI_RC_ERR_FAILED, msg);
+      mlogf(M_ERROR,M_SHOW,"%s",errstr);
+      _SFCB_TRACE(1, (errstr));
+      resp = errorCharsResp(CMPI_RC_ERR_FAILED, errstr);
+	  free (errstr); 
+	  errstr = NULL; 
    }
    
    else {
-      _SFCB_TRACE(1, ("--- Provider request for %s %p %x",
+      _SFCB_TRACE(1, ("--- Provider request for op:%s pInfo:%p prov:%x",
         opsName[req->operation],pInfo,req->provId));
 
       if (req->flags & FL_chunked) requestor=parms->requestor;
@@ -2842,12 +2695,11 @@ static void *processProviderInvocationRequestsThread(void *prms)
 	if (req->operation == OPS_LoadProvider && resp->rc == 2)
 	  exit(-1);
       }
+      free(resp);
    }  
     
    tool_mm_flush();
    
-   free(resp);
-
    if (pInfo && idleThreadStartHandled==0) {
       if (idleThreadStartHandled==0 && req->operation != OPS_PingProvider) {
          if (pInfo->unload==0) {
@@ -2868,35 +2720,34 @@ static void *processProviderInvocationRequestsThread(void *prms)
    if ((req->options & BRH_Internal) ==0) close(abs(parms->requestor));
    free(parms);
    free(req);
-  
+
    _SFCB_RETURN(NULL);
 }
 
 int pauseProvider(char *name)
 {
    int rc=0;
-   char *n;
-  
+   char *n, *m;
    if (noProvPause) return 0;
-   if (provPauseStr==NULL) return 0;
+   if (provPauseStr == NULL) return 0;
    else {
-      if (provPauseStr) {
-         char *p;
-         p=provPauseStr=strdup(provPauseStr);
-         while (*p) { *p=tolower(*p); p++; }
-      }
-   }   
-   if (provPauseStr) {
+      char *p;
+      p=m=strdup(provPauseStr);
+      while (*p) { *p=tolower(*p); p++; }
+   }
+   if (name) {
       char *p;
       int l=strlen(name);
-      p=n=strdup(name);      
+      p=n=strdup(name);
       while (*p) { *p=tolower(*p); p++; }
-      if ((p=strstr(provPauseStr,n))!=NULL) {
-         if ((p==provPauseStr || *(p-1)==',') && (p[l]==',' || p[l]==0)) rc=1;
+      if ((p=strstr(m,n))!=NULL) {
+         if ((p == m || *(p-1) == ',') && (p[l]==',' || p[l]==0)) rc=1;
       }
-      free(p);
+      free(m);
+      free(n);
       return rc;
-  }
+   }
+   free(m);
    noProvPause=1;
    return 0;
 }
@@ -2917,7 +2768,7 @@ void processProviderInvocationRequests(char *name)
 
    debugMode=pauseProvider(name);
    for (;;) {
-      _SFCB_TRACE(1, ("--- Waiting for provider request to %d-%lu",
+      _SFCB_TRACE(1, ("--- Waiting for provider request to R%d-%lu",
                    providerSockets.receive,getInode(providerSockets.receive)));
       parms = (Parms *) malloc(sizeof(*parms));
       
@@ -2927,7 +2778,7 @@ void processProviderInvocationRequests(char *name)
 	int debug_break = 0;
          if (rc!=0)mlogf(M_ERROR,M_SHOW,"oops\n");               
 
-         _SFCB_TRACE(1, ("--- Got something %d-%p on %d-%lu", 
+         _SFCB_TRACE(1, ("--- Got something op:%d-prov:%p on R%d-%lu", 
             parms->req->operation,parms->req->provId,
             providerSockets.receive,getInode(providerSockets.receive)));
 
