@@ -1,6 +1,6 @@
 
 /*
- * $Id: providerMgr.c,v 1.50 2008/04/16 17:41:29 gdread Exp $
+ * $Id: providerMgr.c,v 1.61 2009/04/17 21:42:27 smswehla Exp $
  *
  * © Copyright IBM Corp. 2005, 2007
  *
@@ -120,22 +120,6 @@ static void notSupported(int *requestor, OperationHdr * req)
    free(req);
 }
 
-/*
-static void handleSigterm(int sig)
-{
-   mlogf(M_ERROR,M_SHOW, "%s: exiting due to signal %d\n", "provider", sig);
-   exit(1);
-}
-
-static void handleSigSegv(int sig)
-{
-   mlogf(M_ERROR,M_SHOW, "()%d): exiting due to a SIGSEGV signal %d - %s(%d)\n",
-           currentProc, sig, __FILE__, __LINE__);
-   abort();
-}
-*/
-
-
 /* -------------
  * ---
  *      Instance Provider support
@@ -194,6 +178,12 @@ int nameSpaceOk(ProviderInfo *info, char *nameSpace)
    _SFCB_RETURN(0);
 }
 
+/*
+ * Find the provider for className residing in nameSpace
+ * If none is found for that specific class, try the parent class
+ * If still none, return DefaultProvider
+ *
+ */
 static ProviderInfo *lookupProvider(long type, char *className, char *nameSpace, 
       CMPIStatus *st)
 {
@@ -371,7 +361,6 @@ static void lookupProviderList(long type, int *requestor, OperationHdr * req)
    unsigned long count,n;
    ProviderInfo *info;
    int dmy = 0, rc,indFound=0;
-   char *msg = NULL;
    CMPIStatus st = {CMPI_RC_OK, NULL};
 
    providers = lookupProviders(type,className,nameSpace,&st);
@@ -383,7 +372,7 @@ static void lookupProviderList(long type, int *requestor, OperationHdr * req)
       for (info = providers->ft->getFirst(providers); info;
            info = providers->ft->getNext(providers), n--) {
          if (info->type!=FORCE_PROVIDER_NOTFOUND &&
-             (rc = forkProvider(info, req, &msg)) == CMPI_RC_OK) {
+             (rc = forkProvider(info, req, NULL)) == CMPI_RC_OK) {
             _SFCB_TRACE(1,("--- responding with  %s %p %d",info->providerName,info,count));
             spSendCtlResult(requestor, &info->providerSockets.send,
                 MSG_X_PROVIDER, count--, getProvIds(info).ids, req->options);
@@ -395,7 +384,6 @@ static void lookupProviderList(long type, int *requestor, OperationHdr * req)
             };
             spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
                NULL, req->options);
-            if (msg) free(msg);
 	    break;
          }
       }
@@ -414,7 +402,6 @@ static void findProvider(long type, int *requestor, OperationHdr * req)
    char *className = (char *) req->className.data;
    char *nameSpace = (char *) req->nameSpace.data;
    ProviderInfo *info;
-   char *msg=NULL;
    int rc;
    CMPIStatus st;
 
@@ -422,14 +409,13 @@ static void findProvider(long type, int *requestor, OperationHdr * req)
    
    if ((info = lookupProvider(type,className,nameSpace,&st)) != NULL) {
       if (info->type!=FORCE_PROVIDER_NOTFOUND &&
-          (rc = forkProvider(info, req, &msg)) == CMPI_RC_OK) {
+          (rc = forkProvider(info, req, NULL)) == CMPI_RC_OK) {
          spSendCtlResult(requestor, &info->providerSockets.send, MSG_X_PROVIDER,
             0, getProvIds(info).ids, req->options);
       }                   
       else {
          spSendCtlResult(requestor, &sfcbSockets.send, MSG_X_PROVIDER_NOT_FOUND, 0, 
             NULL, req->options);
-         if (msg) free(msg);
       }
    }
    else {
@@ -599,7 +585,6 @@ static void assocProviderList(int *requestor, OperationHdr * req)
    long count = 0;
    ProviderInfo *info;
    int dmy = 0, rc;
-   char *msg=NULL;
 
    if (className==NULL || *className == 0)
       className = "$ASSOCCLASSES$";
@@ -611,7 +596,7 @@ static void assocProviderList(int *requestor, OperationHdr * req)
          for (info = providers->ft->getFirst(providers); info;
               info = providers->ft->getNext(providers)) {
             if (info->type!=FORCE_PROVIDER_NOTFOUND &&
-                (rc = forkProvider(info, req, &msg)) == CMPI_RC_OK) {
+                (rc = forkProvider(info, req, NULL)) == CMPI_RC_OK) {
                _SFCB_TRACE(1,("--- responding with  %s %p %d",info->providerName,info,count));
                spSendCtlResult(requestor, &info->providerSockets.send,
                   MSG_X_PROVIDER, count--, getProvIds(info).ids, req->options);
@@ -619,10 +604,8 @@ static void assocProviderList(int *requestor, OperationHdr * req)
             else {
                spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND,
                   0, NULL, req->options);
-               if (msg) free(msg);
 	       break;
             }
-
          }
       }
       else {
@@ -634,7 +617,7 @@ static void assocProviderList(int *requestor, OperationHdr * req)
       /* When there is no provider for an assocClass we do not want to produce
        * an error message. So we return the default provider and expect it
        * to produce a nice and empty result */
-      if((rc = forkProvider(defaultProvInfoPtr, req, &msg)) == CMPI_RC_OK) {
+      if((rc = forkProvider(defaultProvInfoPtr, req, NULL)) == CMPI_RC_OK) {
          _SFCB_TRACE(1,("--- responding with  %s %p %d",
                         defaultProvInfoPtr->providerName,
                         defaultProvInfoPtr,count));
@@ -642,6 +625,14 @@ static void assocProviderList(int *requestor, OperationHdr * req)
          MSG_X_PROVIDER, count--, getProvIds(defaultProvInfoPtr).ids,
                                                        req->options);
       }
+     else 
+       {
+	 /* Oops, even creation of the default provider failed */
+	 mlogf(M_ERROR,M_SHOW,"--- forkProvider failed for defaultProvider\n");
+	 /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+	  NULL, req->options); */
+	 _SFCB_ABORT();
+       }     
    }
    _SFCB_EXIT();
 }
@@ -705,10 +696,17 @@ static ProviderInfo *getMethodProvider(char *className, char *nameSpace)
 
 static void classProvider(int *requestor, OperationHdr * req)
 {
-   char *msg;
-
+  int rc;
    _SFCB_ENTER(TRACE_PROVIDERMGR, "classProvider");
-   forkProvider(classProvInfoPtr, req, &msg);
+   rc = forkProvider(classProvInfoPtr, req, NULL);
+   if (rc != CMPI_RC_OK)
+    {
+      mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in classProvider\n");
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
+      _SFCB_ABORT();
+    }
+  
    _SFCB_TRACE(1,("--- result %d-%lu to with %d-%lu",
       *requestor,getInode(*requestor),
       classProvInfoPtr->providerSockets.send,
@@ -716,16 +714,21 @@ static void classProvider(int *requestor, OperationHdr * req)
    
    spSendCtlResult(requestor, &classProvInfoPtr->providerSockets.send, MSG_X_PROVIDER,
       0, getProvIds(classProvInfoPtr).ids, req->options);
-                                   
    _SFCB_EXIT();
 }
 
 static void qualiProvider(int *requestor, OperationHdr * req)
 {
-   char *msg;
-   
+   int rc;
    _SFCB_ENTER(TRACE_PROVIDERMGR, "qualiProvider");
-   forkProvider(qualiProvInfoPtr, req, &msg);
+   rc = forkProvider(qualiProvInfoPtr, req, NULL);
+   if (rc != CMPI_RC_OK)
+    {
+      mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in qualiProvider\n");
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
+      _SFCB_ABORT();
+    }
    _SFCB_TRACE(1,("--- result %d-%lu to with %d-%lu",   
       *requestor,getInode(*requestor),
       qualiProvInfoPtr->providerSockets.send,
@@ -738,7 +741,6 @@ static void qualiProvider(int *requestor, OperationHdr * req)
 
 static void methProvider(int *requestor, OperationHdr * req)
 {
-   char *msg=NULL;
    int rc;
    char *className = (char *) req->className.data;
    char *nameSpace = (char *) req->nameSpace.data;
@@ -748,16 +750,21 @@ static void methProvider(int *requestor, OperationHdr * req)
    if (strcmp(className, "$ClassProvider$") == 0)
       classProvider(requestor, req);
    else if ((info = getMethodProvider(className,nameSpace)) != NULL) {
+      rc = CMPI_RC_OK;
       if (info->type!=FORCE_PROVIDER_NOTFOUND &&
-          (rc = forkProvider(info, req, &msg)) == CMPI_RC_OK) {
+          (rc = forkProvider(info, req, NULL)) == CMPI_RC_OK) {
          _SFCB_TRACE(1,("--- responding with  %s %p",info->providerName,info));
          spSendCtlResult(requestor, &info->providerSockets.send, MSG_X_PROVIDER,
             0, getProvIds(info).ids, req->options);
       }                   
       else {
-         spSendCtlResult(requestor, &sfcbSockets.send, MSG_X_PROVIDER_NOT_FOUND,
+	if (rc != CMPI_RC_OK)
+	  {
+	    mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in methProvider\n");
+	  }
+	
+	  spSendCtlResult(requestor, &sfcbSockets.send, MSG_X_PROVIDER_NOT_FOUND,
             0, NULL, req->options);
-         if (msg) free(msg);
       }
    }
    else
@@ -766,6 +773,7 @@ static void methProvider(int *requestor, OperationHdr * req)
    _SFCB_EXIT();
 }
 
+
 static int _methProvider(BinRequestContext * ctx, OperationHdr * req)
 {
    _SFCB_ENTER(TRACE_PROVIDERMGR, "_methProvider");
@@ -773,35 +781,51 @@ static int _methProvider(BinRequestContext * ctx, OperationHdr * req)
    char *nameSpace = (char *) req->nameSpace.data;
    ProviderInfo *info;
    int rc;
-   char *msg;
 
    ctx->chunkedMode=ctx->xmlAs=0;
    if (strcmp(className, "$ClassProvider$") == 0) {
-      forkProvider(classProvInfoPtr, req, &msg);
+      rc = forkProvider(classProvInfoPtr, req, NULL);
+      if (rc != CMPI_RC_OK)
+        {
+	  mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in _methProvider (%s)\n", className);
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
+	  _SFCB_ABORT();
+        }
       ctx->provA.ids = getProvIds(classProvInfoPtr);
       ctx->provA.socket = classProvInfoPtr->providerSockets.send;
       ctx->pAs=NULL;
       _SFCB_RETURN(MSG_X_PROVIDER); 
    }
    else if (strcmp(className, "$InterOpProvider$") == 0) {
-      forkProvider(interOpProvInfoPtr, req, &msg);
+      rc = forkProvider(interOpProvInfoPtr, req, NULL);
+      if (rc != CMPI_RC_OK)
+        {
+	  mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in _methProvider (%s)\n", className);
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
+	  _SFCB_ABORT();
+        }
       ctx->provA.ids = getProvIds(interOpProvInfoPtr);
       ctx->provA.socket = interOpProvInfoPtr->providerSockets.send;
       ctx->pAs=NULL;
       _SFCB_RETURN(MSG_X_PROVIDER);
    }
    else if ((info = getMethodProvider(className,nameSpace)) != NULL) {
-      if ((rc = forkProvider(info, req, &msg)) == CMPI_RC_OK) {
+      if ((rc = forkProvider(info, req, NULL)) == CMPI_RC_OK) {
          ctx->provA.ids = getProvIds(info);
          ctx->provA.socket = info->providerSockets.send;
          ctx->pAs=NULL; 
          _SFCB_RETURN(MSG_X_PROVIDER);
       }
-      else {
-         free(msg);
+      else { 
+	 mlogf(M_ERROR,M_SHOW,"--- _methProvider NOT FOUND\n");
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
          _SFCB_RETURN(MSG_X_PROVIDER_NOT_FOUND);
       }
    }
+   mlogf(M_ERROR,M_SHOW,"--- _methProvider INVALID\n");
    _SFCB_RETURN(MSG_X_INVALID_CLASS);
 }
 
@@ -857,7 +881,7 @@ void processProviderMgrRequests()
    _SFCB_ENTER(TRACE_PROVIDERMGR, "processProviderMgrRequests");
 
    startUpProvider("root/interop","$ClassProvider$");
-   
+
 #ifdef SFCB_INCL_INDICATION_SUPPORT
    if (interOpProvInfoPtr!=forceNoProvInfoPtr) {
      sleep(2);
@@ -905,20 +929,38 @@ void processProviderMgrRequests()
 }
 
 void closeProviderContext(BinRequestContext * ctx)
-{ 
+{
+   int i = 0;
+    _SFCB_ENTER(TRACE_PROVIDERMGR, "closeProviderContext");
+   for(i = 0; i < ctx->pCount; i++) {
+     semAcquire(sfcbSem,PROV_GUARD(ctx->pAs[i].ids.procId));
+     if(semGetValue(sfcbSem, PROV_INUSE(ctx->pAs[i].ids.procId)) != 0) {
+       semAcquire(sfcbSem,PROV_INUSE(ctx->pAs[i].ids.procId));
+     } else {
+       fprintf(stderr, "--- closeProviderContext not touching sem %d; already zero\n", PROV_INUSE(ctx->pAs[i].ids.procId));
+     }
+     semRelease(sfcbSem,PROV_GUARD(ctx->pAs[i].ids.procId));
+   }
    if (ctx->pAs) free(ctx->pAs);
 }
 
 static void setInuseSem(void *id)
 {
    ProvIds ids;
+   key_t semKey;
 
-   _SFCB_ENTER(TRACE_PROVIDERMGR, "setInuseSem");
-   ids.ids=id;   
-   semAcquire(sfcbSem,(ids.procId*3)+provProcGuardId+provProcBaseId);
-   semAcquire(sfcbSem,(ids.procId*3)+provProcInuseId+provProcBaseId);
-   semReleaseUnDo(sfcbSem,(ids.procId*3)+provProcInuseId+provProcBaseId);
-   semRelease(sfcbSem,(ids.procId*3)+provProcGuardId+provProcBaseId);
+    _SFCB_ENTER(TRACE_PROVIDERMGR, "setInuseSem");
+   if (sfcbSem < 0) { //Semaphore Not initialized.
+        semKey=ftok(SFCB_BINARY,'S');
+       sfcbSem=semget(semKey,1, 0600);
+   }
+
+   ids.ids=id;
+
+   semAcquire(sfcbSem,PROV_GUARD(ids.procId));
+   //semAcquire(sfcbSem,PROV_INUSE(ids.procId));
+   //semReleaseUnDo(sfcbSem,PROV_INUSE(ids.procId));
+   semRelease(sfcbSem,PROV_GUARD(ids.procId));
    _SFCB_EXIT();
 }
 
@@ -926,7 +968,7 @@ static void setInuseSem(void *id)
 int getProviderContext(BinRequestContext * ctx, OperationHdr * ohdr)
 {
    unsigned long int l;
-   int rc, i, x;
+   int rc=0, i, x;
    char *buf;
    ProvAddr *as;
    ComSockets sockets;
@@ -954,9 +996,17 @@ int getProviderContext(BinRequestContext * ctx, OperationHdr * ohdr)
    
    _SFCB_TRACE(1,("--- Sending mgr request - to %d from %d", sfcbSockets.send,
                 sockets.send)); 
-   spSendReq(&sfcbSockets.send, &sockets.send, buf, l, localMode);   
+   rc = spSendReq(&sfcbSockets.send, &sockets.send, buf, l, localMode);   
    free(buf);
 
+   if (rc < 0) {
+      mlogf(M_ERROR,M_SHOW,"--- spSendReq/spSendMsg failed to send on %d (%d)\n", sfcbSockets.send, rc);
+      ctx->rc = rc;
+      close (sockets.send);
+      close (sockets.receive);
+      return rc;
+   }
+
    _SFCB_TRACE(1, ("--- Sending mgr request done"));
    
    ctx->rc = spRecvCtlResult(&sockets.receive, &ctx->provA.socket, &ctx->provA.ids.ids, &l);
@@ -1297,7 +1347,7 @@ CMPIConstClass *getConstClass(const char *ns, const char *cn)
    
    unlockUpCall(Broker);
 
-   free(resp);
+   if(resp) free(resp);
    if(!localMode){
   	close(binCtx.provA.socket);
    }
@@ -1317,7 +1367,6 @@ static CMPIConstClass *_getConstClass(const char *ns, const char *cn, CMPIStatus
    BinRequestContext binCtx;
    OperationHdr req = { OPS_GetClass, 2 };
    int irc;
-   char *msg;
    
    path = NewCMPIObjectPath(ns, cn, st);
    sreq.objectPath = setObjectPathMsgSegment(path);
@@ -1326,8 +1375,15 @@ static CMPIConstClass *_getConstClass(const char *ns, const char *cn, CMPIStatus
    req.nameSpace = setCharsMsgSegment((char *) ns);
    req.className = setCharsMsgSegment((char *) cn);
 
-   forkProvider(classProvInfoPtr, &req, &msg);
-   
+   irc = forkProvider(classProvInfoPtr, &req, NULL);
+   if (irc != CMPI_RC_OK) 
+    {
+      mlogf(M_ERROR,M_SHOW,"--- forkProvider failed in _getConstClass(%s:%s)\n", ns, cn);
+      /*      spSendCtlResult(requestor, &dmy, MSG_X_PROVIDER_NOT_FOUND, 0,
+               NULL, req->options); */
+      _SFCB_ABORT();
+    }
+  
    memset(&binCtx,0,sizeof(BinRequestContext));
    binCtx.oHdr = &req;
    binCtx.bHdr = &sreq.hdr;
@@ -1336,22 +1392,20 @@ static CMPIConstClass *_getConstClass(const char *ns, const char *cn, CMPIStatus
    binCtx.chunkedMode=binCtx.xmlAs=binCtx.noResp=0;
 
    binCtx.provA.socket = classProvInfoPtr->providerSockets.send;
-   irc = MSG_X_PROVIDER;
 
-   if (irc) {
-      _SFCB_TRACE(1, ("--- Invoking ClassProvider for %s %s",ns,cn));
-      resp = invokeProvider(&binCtx);
-      resp->rc--;
-      st->rc=resp->rc;
-      if (resp->rc == CMPI_RC_OK) {
-         ccl = relocateSerializedConstClass(resp->object[0].data);
-         ccl = ccl->ft->clone(ccl, NULL);
-      }
+   _SFCB_TRACE(1, ("--- Invoking ClassProvider for %s %s",ns,cn));
+   resp = invokeProvider(&binCtx);
+   resp->rc--;
+   st->rc=resp->rc;
+   if (resp->rc == CMPI_RC_OK) {
+      ccl = relocateSerializedConstClass(resp->object[0].data);
+      ccl = ccl->ft->clone(ccl, NULL);
    }
+
    _SFCB_TRACE(1, ("--- Invoking ClassProvider for %s %s rc: %d",ns,cn,resp->rc));
    
    path->ft->release(path);
-   free(resp);
+   if(resp) free(resp);
 
    _SFCB_RETURN(ccl);
 }
@@ -1424,7 +1478,7 @@ int isChild(const char *ns, const char *parent, const char* child)
 
    irc = _methProvider(&binCtx, &req);
 
-   if (irc) {
+   if (irc == MSG_X_PROVIDER) {
       localInvokeMethod(&binCtx, path, "ischild", in, &out, &rc,0);
       irc=(rc.rc==CMPI_RC_OK);
    }
@@ -1454,7 +1508,7 @@ static int startUpProvider(const char* ns, const char *name)
 
    irc = _methProvider(&binCtx, &req);
 
-   if (irc) {
+   if (irc == MSG_X_PROVIDER) {
       localInvokeMethod(&binCtx, path, "_startup", in, &out, &rc, 1);
       irc=(rc.rc==CMPI_RC_OK);
    }
@@ -1490,7 +1544,7 @@ static UtilList *_getConstClassChildren(const char *ns, const char *cn)
 
    irc = _methProvider(&binCtx, &req);
 
-   if (irc) {
+   if (irc == MSG_X_PROVIDER) {
       data = localInvokeMethod(&binCtx, path, "getchildren", in, &out, &rc, 0);
       if (out) {
          ar = CMGetArg(out, "children", &rc).value.array;
@@ -1532,7 +1586,7 @@ static UtilList *_getAssocClassNames(const char *ns)
    memset(&binCtx,0,sizeof(BinRequestContext));
    irc = _methProvider(&binCtx, &req);
 
-   if (irc) {
+   if (irc == MSG_X_PROVIDER) {
       data = localInvokeMethod(&binCtx, path, "getassocs", in, &out, &rc,0);
       if (out) {
          ar = CMGetArg(out, "assocs", &rc).value.array;
