@@ -43,6 +43,7 @@
 #include "control.h"
 
 #include <getopt.h>
+#include <syslog.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -53,6 +54,11 @@
 static int startSLP = 1;
 #endif
 
+#ifdef LOCAL_CONNECT_ONLY_ENABLE
+// from httpAdapter.c
+int sfcBrokerPid=0;
+#endif // LOCAL_CONNECT_ONLY_ENABLE
+
 //#define USE_THREADS
 
 // for use by thread in startHttpd, startDbpd
@@ -221,12 +227,12 @@ static void stopBroker(void *p)
    stopLocalConnectServer();
    
    for(;;) {
-      waitTime.tv_sec=time(NULL)+5;
-      waitTime.tv_nsec=0;
 
       if (adaptersStopped==0) {
          pthread_mutex_lock(&sdMtx);
-         if (sa==0) mlogf(M_INFO,M_SHOW,"--- Stopping adapters\n");
+         waitTime.tv_sec=time(NULL)+5;
+         waitTime.tv_nsec=0;
+         if (sa==0) fprintf(stderr,"--- Stopping adapters\n");
          sa++;
          if (stopNextAdapter()) {
             rc=pthread_cond_timedwait(&sdCnd,&sdMtx,&waitTime);
@@ -240,7 +246,9 @@ static void stopBroker(void *p)
       
       if (adaptersStopped) {
          pthread_mutex_lock(&sdMtx);
-         if (sp==0) mlogf(M_INFO,M_SHOW,"--- Stopping providers\n");
+         waitTime.tv_sec=time(NULL)+5;
+         waitTime.tv_nsec=0;
+         if (sp==0)  fprintf(stderr,"--- Stopping providers\n");
          sp++;
          if (stopNextProc()) {
             rc=pthread_cond_timedwait(&sdCnd,&sdMtx,&waitTime);
@@ -260,9 +268,8 @@ static void stopBroker(void *p)
    
    if (restartBroker) {
       char *emsg=strerror(errno);
-      mlogf(M_INFO,M_SHOW,"---\n");   
       execvp("sfcbd",restartArgv);
-      mlogf(M_ERROR,M_SHOW,"--- execv for restart problem: %s\n",emsg);
+      fprintf(stderr,"--- execv for restart problem: %s\n",emsg);
       abort();
    }
 
@@ -298,7 +305,7 @@ static void handleSigquit(int sig)
    pthread_attr_t tattr;
    
    if (sfcBrokerPid==currentProc) {    
-      mlogf(M_INFO,M_SHOW, "--- Winding down %s\n", processName);
+      fprintf(stderr,"--- Winding down %s\n", processName);
       pthread_attr_init(&tattr);
       pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);      
       pthread_create(&t, &tattr, (void *(*)(void *))stopBroker,NULL);
@@ -314,7 +321,7 @@ static void handleSigHup(int sig)
   
    if (sfcBrokerPid==currentProc) {    
       restartBroker=1;
-      mlogf(M_INFO,M_SHOW, "--- Restarting %s\n", processName);
+      fprintf(stderr,"--- Restarting %s\n", processName);
       pthread_attr_init(&tattr);
       pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);      
       pthread_create(&t, &tattr, (void *(*)(void *))stopBroker,NULL);
@@ -348,7 +355,7 @@ static void handleSigChld(int sig)
 //         mlogf(M_INFO,M_SHOW,"sigchild %d\n",pid);
         if (testStartedAdapter(pid,&left)) { 
             if (left==0) {
-               mlogf(M_INFO,M_SHOW,"--- Adapters stopped\n");
+               fprintf(stderr,"--- Adapters stopped\n");
                adaptersStopped=1;
             }   
             pthread_attr_init(&tattr);
@@ -357,7 +364,7 @@ static void handleSigChld(int sig)
          }
          else if (testStartedProc(pid,&left)) {
             if (left==0) {
-               mlogf(M_INFO,M_SHOW,"--- Providers stopped\n");
+               fprintf(stderr,"--- Providers stopped\n");
                providersStopped=1;
             }   
             pthread_attr_init(&tattr);
@@ -374,7 +381,7 @@ static void handleSigterm(int sig)
 {
 
    if (!terminating) {
-      mlogf(M_ERROR,M_SHOW, "--- %s - %d exiting due to signal %d\n", processName, currentProc, sig);
+      fprintf(stderr,"--- %s - %d exiting due to signal %d\n", processName, currentProc, sig);
       dumpTiming(currentProc);
    }   
    terminating=1; 
@@ -386,18 +393,19 @@ static void handleSigterm(int sig)
 
 static void handleSigSegv(int sig)
 {
-   mlogf(M_ERROR,M_SHOW, "-#- %s - %d exiting due to a SIGSEGV signal\n",
+   fprintf(stderr,"-#- %s - %d exiting due to a SIGSEGV signal\n",
            processName, currentProc);
 }
 /*
 static void handleSigAbort(int sig)
 {
-   mlogf(M_INFO,M_SHOW, "%s: exiting due to a SIGABRT signal - %d\n", processName, currentProc);
+   fprintf(stderr,"%s: exiting due to a SIGABRT signal - %d\n", processName, currentProc);
    kill(0, SIGTERM);
 }
 */
 
 
+#ifndef LOCAL_CONNECT_ONLY_ENABLE
 #ifdef USE_THREADS
 void* startHttpThread(void* params) {
   struct adapterThreadParams* p = (struct adapterThreadParams*)params;
@@ -448,6 +456,8 @@ static int startHttpd(int argc, char *argv[], int sslMode)
    return 0;
 }
 
+#endif // LOCAL_CONNECT_ONLY_ENABLE
+
 #ifdef HAVE_JDBC
 
 extern int dbpDaemon(int argc, char *argv[], int sslMode, int sfcbPid);
@@ -500,6 +510,9 @@ static void usage(int status)
 		" -d, --daemon                    run in the background",
 		" -h, --help                      display this message and exit",
 		" -k, --color-trace               color the trace output of each process",
+		" -l, --syslog-level=<LOGLEVEL>   specify the level for syslog",
+		"                                 LOGLEVEL can be LOG_INFO, LOG_DEBUG, or LOG_ERR",
+		"                                 LOG_ERR is the default",
 		" -s, --collect-stats             turn on runtime statistics collecting",
 		" -t, --trace-components=<N|?>    activate component-level tracing messages where",
 		"                                 N is an OR-ed bitmask integer defining the",
@@ -533,7 +546,11 @@ int main(int argc, char *argv[])
    int c, i;
    long tmask = 0, sslMode=0,sslOMode=0, tracelevel=0;
    char * tracefile = NULL;
+#ifdef HAVE_UDS
+   int enableUds=0;
+#endif
    int enableHttp=0,enableHttps=0,useChunking=0,doBa=0,enableInterOp=0,httpLocalOnly=0;
+   int syslogLevel=LOG_ERR;
    long dSockets,sSockets,pSockets;
    char *pauseStr;
 
@@ -550,7 +567,6 @@ int main(int argc, char *argv[])
    restartArgc=argc;
    restartArgv=argv;
 
-   startLogging("sfcb");
 
    exFlags = 0;
 
@@ -561,12 +577,13 @@ int main(int argc, char *argv[])
 	   { "help",             no_argument,       0,        'h' },
 	   { "color-trace",      no_argument,       0,        'k' },
 	   { "collect-stats",    no_argument,       0,        's' },
+	   { "syslog-level",     required_argument, 0,        'l' },
 	   { "trace-components", required_argument, 0,        't' },
 	   { "version",          no_argument,       0,        'v' },
 	   { 0, 0, 0, 0 }
        };
 
-   while ((c = getopt_long(argc, argv, "c:dhkst:v", long_options, 0)) != -1)
+   while ((c = getopt_long(argc, argv, "c:dhkst:vl:", long_options, 0)) != -1)
    {
        switch(c)
        {
@@ -579,6 +596,7 @@ int main(int argc, char *argv[])
 
 	   case 'd':
 	       daemon(0, 0);
+	       currentProc=sfcBrokerPid=getpid(); /* req. on some systems */
 	       break;
 
 	   case 'h':
@@ -610,10 +628,24 @@ int main(int argc, char *argv[])
 	   case 'v':
 	       version();
 
+	   case 'l':
+            if (strcmp(optarg,"LOG_ERR")==0) {
+                syslogLevel=LOG_ERR;
+            } else if (strcmp(optarg,"LOG_INFO")==0) {
+                syslogLevel=LOG_INFO;
+            } else if (strcmp(optarg,"LOG_DEBUG")==0) {
+                syslogLevel=LOG_DEBUG;
+            } else {
+                fprintf(stderr,"Invalid value for syslog-level.\n");
+	            usage(3);
+            }
+            break;
+        
 	   default:
 	       usage(3);
        }
    }
+   startLogging("sfcb",syslogLevel);
 
    mlogf(M_INFO,M_SHOW,"--- %s V" sfcHttpDaemonVersion " started - %d\n", name, currentProc);
 
@@ -656,13 +688,22 @@ int main(int argc, char *argv[])
       
    if (getControlBool("enableHttp", &enableHttp))
       enableHttp=1;
+
+#ifdef HAVE_UDS
+   if (getControlBool("enableUds", &enableUds))
+      enableUds=1;
+#endif
       
 #if defined USE_SSL
    if (getControlBool("enableHttps", &enableHttps))
       enableHttps=0;
 
    sslMode=enableHttps;
+#ifdef HAVE_UDS
+   sslOMode=sslMode & !enableHttp & !enableUds;
+#else
    sslOMode=sslMode & !enableHttp;
+#endif
 #else
    mlogf(M_INFO,M_SHOW,"--- SSL not configured\n");
    enableHttps=0;
@@ -710,12 +751,12 @@ int main(int argc, char *argv[])
    else
        exFlags = exFlags | 2;
 
-   if ((enableInterOp && pSockets < 2 ) || pSockets < 1) {
+   if ((enableInterOp && pSockets < 4 ) || pSockets < 3) {
      /* adjusting provider number */
      if (enableInterOp) {
-       pSockets = 2;
+       pSockets = 4;
      } else {
-       pSockets = 1;
+       pSockets = 3;
      }
      mlogf(M_INFO,M_SHOW,
 	   "--- Max provider process number adjusted to %d\n", pSockets);
@@ -740,12 +781,14 @@ int main(int argc, char *argv[])
 
    startLocalConnectServer();
    
+#ifndef LOCAL_CONNECT_ONLY_ENABLE
    if (startHttp) {
       if (sslMode)
          startHttpd(argc, argv,1);
       if (!sslOMode)
          startHttpd(argc, argv,0);
    }
+#endif // LOCAL_CONNECT_ONLY_ENABLE
    
 #ifdef HAVE_JDBC
    //Start dbProtocol-Daemon
