@@ -1,6 +1,6 @@
 
 /*
- * $Id: support.c,v 1.26 2008/01/24 21:56:29 buccella Exp $
+ * $Id: support.c,v 1.33 2009/03/10 19:18:19 buccella Exp $
  *
  *  © Copyright IBM Corp. 2005, 2007
  *
@@ -37,6 +37,7 @@
 #include "native.h"
 #include "trace.h"
 #include "config.h"
+#include <pthread.h>
 
 #ifdef SFCB_IX86
 #define SFCB_ASM(x) asm(x)
@@ -51,7 +52,7 @@ int localClientMode=0;        /**< flag determining whether local client connect
 void *loadLibib(const char *libname)
 {
    char filename[255];
-   sprintf(filename, "lib%s.so", libname);
+   snprintf(filename, 255, "lib%s.so", libname);
    return dlopen(filename, RTLD_LAZY);
 }
 
@@ -60,7 +61,7 @@ static void *getGenericEntryPoint(void *library, const char *ptype)
 {
    char entry_point[255];
    void *sym;
-   sprintf(entry_point, "_Generic_Create_%sMI", ptype);
+   snprintf(entry_point, 255, "_Generic_Create_%sMI", ptype);
    sym = dlsym(library, entry_point);
    return sym;
 }
@@ -71,7 +72,7 @@ static void *getFixedEntryPoint(const char *provider,
 {
    char entry_point[255];
    void *sym;
-   sprintf(entry_point, "%s_Create_%sMI", provider, ptype);
+   snprintf(entry_point, 255, "%s_Create_%sMI", provider, ptype);
    sym = dlsym(library, entry_point);
    return sym;
 }
@@ -79,15 +80,19 @@ static void *getFixedEntryPoint(const char *provider,
 
 typedef CMPIInstanceMI *(*GENERIC_InstanceMI) (CMPIBroker * broker,
                                                CMPIContext * ctx,
-                                               const char *provider);
+                                               const char * provider,
+					       CMPIStatus * status);
 typedef CMPIInstanceMI *(*FIXED_InstanceMI) (CMPIBroker * broker,
-                                             CMPIContext * ctx);
+                                             CMPIContext * ctx,
+					     CMPIStatus * status);
 
 
 CMPIInstanceMI *loadInstanceMI(const char *provider, void *library,
-                                     CMPIBroker * broker, CMPIContext * ctx)
+                                     CMPIBroker * broker, CMPIContext * ctx, 
+                                     CMPIStatus * status)
 {
    CMPIInstanceMI *mi; 
+
    _SFCB_ENTER(TRACE_PROVIDERDRV, "loadInstanceMI");
    
    GENERIC_InstanceMI g = (GENERIC_InstanceMI) 
@@ -97,32 +102,38 @@ CMPIInstanceMI *loadInstanceMI(const char *provider, void *library,
          getFixedEntryPoint(provider, library, "Instance");
       if (f == NULL) _SFCB_RETURN(NULL);
       if (broker) {
-         mi=(f) (broker, ctx);
+	mi=(f) (broker, ctx, status);
+	if (mi && status->rc == CMPI_RC_OK)
          _SFCB_RETURN(mi);
       }   
-      _SFCB_RETURN((CMPIInstanceMI*)(void*)-1);
+      _SFCB_RETURN(NULL); 
    }
    if (broker) {
-      mi=(g) (broker, ctx, provider);
-      _SFCB_RETURN(mi);
+     mi=(g) (broker, ctx, provider, status);
+     if (mi && status->rc == CMPI_RC_OK)
+        _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIInstanceMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 
 typedef CMPIAssociationMI *(*GENERIC_AssociationMI) (CMPIBroker * broker,
                                                      CMPIContext * ctx,
-                                                     const char *provider);
+                                                     const char * provider,
+						     CMPIStatus * status);
 typedef CMPIAssociationMI *(*FIXED_AssociationMI) (CMPIBroker * broker,
-                                                   CMPIContext * ctx);
+                                                   CMPIContext * ctx,
+						   CMPIStatus * status);
 
 
 CMPIAssociationMI *loadAssociationMI(const char *provider,
                                            void *library,
                                            CMPIBroker * broker,
-                                           CMPIContext * ctx)
+                                           CMPIContext * ctx,
+                                           CMPIStatus * status)
 {
    CMPIAssociationMI *mi; 
+
    _SFCB_ENTER(TRACE_PROVIDERDRV, "loadAssociationMI");
    
    GENERIC_AssociationMI g = (GENERIC_AssociationMI) 
@@ -133,30 +144,36 @@ CMPIAssociationMI *loadAssociationMI(const char *provider,
          getFixedEntryPoint(provider, library, "Association");
       if (f == NULL) _SFCB_RETURN(NULL);
       if (broker) {
-         mi=(f) (broker, ctx);
-         _SFCB_RETURN(mi);
+	mi=(f) (broker, ctx, status);
+	if (mi && status->rc == CMPI_RC_OK)
+	  _SFCB_RETURN(mi);
       }   
-      _SFCB_RETURN((CMPIAssociationMI*)(void*)-1);
+      _SFCB_RETURN(NULL); 
    }
    if (broker) {
-      mi=(g) (broker, ctx, provider);
-      _SFCB_RETURN(mi);
+     mi=(g) (broker, ctx, provider, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIAssociationMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 
 typedef CMPIMethodMI *(*GENERIC_MethodMI) (CMPIBroker * broker,
                                            CMPIContext * ctelsex,
-                                           const char *provider);
+                                           const char * provider,
+					   CMPIStatus * status);
 typedef CMPIMethodMI *(*FIXED_MethodMI) (CMPIBroker * broker,
-                                         CMPIContext * ctx);
+                                         CMPIContext * ctx,
+					 CMPIStatus * status);
 
 
 CMPIMethodMI *loadMethodMI(const char *provider, void *library,
-                                 CMPIBroker * broker, CMPIContext * ctx)
+                                 CMPIBroker * broker, CMPIContext * ctx,
+                                 CMPIStatus * status)
 {
    CMPIMethodMI *mi; 
+
    _SFCB_ENTER(TRACE_PROVIDERDRV, "loadMethodMI");
    
    GENERIC_MethodMI g =
@@ -166,31 +183,37 @@ CMPIMethodMI *loadMethodMI(const char *provider, void *library,
           (FIXED_MethodMI) getFixedEntryPoint(provider, library,"Method");
       if (f == NULL) _SFCB_RETURN(NULL);
       if (broker) {
-         mi=(f)(broker, ctx);
-         _SFCB_RETURN(mi);
+	mi=(f)(broker, ctx, status);
+	if (mi && status->rc == CMPI_RC_OK)
+	  _SFCB_RETURN(mi);
       }   
-      _SFCB_RETURN((CMPIMethodMI*)(void*)-1);
+      _SFCB_RETURN(NULL); 
    }
    if (broker)  {
-      mi=(g)(broker,ctx, provider);
-      _SFCB_RETURN(mi);
+     mi=(g)(broker,ctx, provider, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIMethodMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 }
 
 
 typedef CMPIPropertyMI *(*GENERIC_PropertyMI) (CMPIBroker * broker,
                                                CMPIContext * ctx,
-                                               const char *provider);
+                                               const char * provider,
+					       CMPIStatus * status);
 typedef CMPIPropertyMI *(*FIXED_PropertyMI) (CMPIBroker * broker,
-                                             CMPIContext * ctx);
+                                             CMPIContext * ctx,
+					     CMPIStatus * status);
 
 CMPIPropertyMI *loadPropertyMI(const char *provider, void *library,
-                                     CMPIBroker * broker, CMPIContext * ctx)
+                                     CMPIBroker * broker, CMPIContext * ctx,
+                                     CMPIStatus * status)
 {
    CMPIPropertyMI *mi; 
 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "loadPropertyMI");
+
    GENERIC_PropertyMI g =
        (GENERIC_PropertyMI) getGenericEntryPoint(library,"Property");
    if (g == NULL) {
@@ -199,34 +222,40 @@ CMPIPropertyMI *loadPropertyMI(const char *provider, void *library,
                                                      "Property");
       if (f == NULL) _SFCB_RETURN(NULL);
       if (broker) {
-         mi=(f)(broker, ctx);
-         _SFCB_RETURN(mi);
+	mi=(f)(broker, ctx, status);
+	if (mi && status->rc == CMPI_RC_OK)
+	  _SFCB_RETURN(mi);
       }   
-      _SFCB_RETURN((CMPIPropertyMI*)(void*)-1);
+      _SFCB_RETURN(NULL); 
    }
 
    if (broker)  {
-      mi=(g)(broker,ctx, provider);
-      _SFCB_RETURN(mi);
+     mi=(g)(broker,ctx, provider, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIPropertyMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 
 typedef CMPIIndicationMI *(*GENERIC_IndicationMI) (CMPIBroker * broker,
                                                    CMPIContext * ctx,
-                                                   const char *provider);
+                                                   const char * provider,
+						   CMPIStatus * status);
 typedef CMPIIndicationMI *(*FIXED_IndicationMI) (CMPIBroker * broker,
-                                                 CMPIContext * ctx);
+                                                 CMPIContext * ctx,
+						 CMPIStatus * status);
 
 
 CMPIIndicationMI *loadIndicationMI(const char *provider,
                                          void *library,
-                                         CMPIBroker * broker, CMPIContext * ctx)
+                                         CMPIBroker * broker, CMPIContext * ctx,
+                                         CMPIStatus * status)
 {
    CMPIIndicationMI *mi; 
 
    _SFCB_ENTER(TRACE_PROVIDERDRV, "loadIndicationMI");
+
    GENERIC_IndicationMI g =
        (GENERIC_IndicationMI) getGenericEntryPoint(library,
                                                         "Indication");
@@ -236,27 +265,31 @@ CMPIIndicationMI *loadIndicationMI(const char *provider,
                                                        "Indication");
       if (f == NULL) _SFCB_RETURN(NULL);
       if (broker) {
-         mi=(f)(broker, ctx);
-         _SFCB_RETURN(mi);
+	mi=(f)(broker, ctx, status);
+	if (mi && status->rc == CMPI_RC_OK)
+	  _SFCB_RETURN(mi);
       }   
-      _SFCB_RETURN((CMPIIndicationMI*)(void*)-1);
+      _SFCB_RETURN(NULL); 
    }
 
    if (broker)  {
-      mi=(g)(broker,ctx, provider);
-      _SFCB_RETURN(mi);
+     mi=(g)(broker,ctx, provider, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIIndicationMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 
 
 typedef CMPIClassMI *(*FIXED_ClassMI) (CMPIBroker * broker,
-                                                 CMPIContext * ctx);
+				       CMPIContext * ctx,
+				       CMPIStatus * status);
 
 CMPIClassMI *loadClassMI(const char *provider,
                                          void *library,
-                                         CMPIBroker * broker, CMPIContext * ctx)
+                                         CMPIBroker * broker, CMPIContext * ctx,
+                                         CMPIStatus * status)
 {
    CMPIClassMI *mi; 
 
@@ -267,18 +300,21 @@ CMPIClassMI *loadClassMI(const char *provider,
    if (f == NULL) _SFCB_RETURN(NULL);
    
    if (broker)  {
-      mi=(f)(broker,ctx);
-      _SFCB_RETURN(mi);
+     mi=(f)(broker,ctx, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIClassMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 typedef CMPIQualifierDeclMI *(*FIXED_QualifierDeclMI) (CMPIBroker * broker,
-                                                 CMPIContext * ctx);
+						       CMPIContext * ctx,
+						       CMPIStatus * status);
 
 CMPIQualifierDeclMI *loadQualifierDeclMI(const char *provider,
                                          void *library,
-                                         CMPIBroker * broker, CMPIContext * ctx)
+                                         CMPIBroker * broker, CMPIContext * ctx,
+                                         CMPIStatus * status)
 {
    CMPIQualifierDeclMI *mi; 
 
@@ -289,10 +325,11 @@ CMPIQualifierDeclMI *loadQualifierDeclMI(const char *provider,
    if (f == NULL) _SFCB_RETURN(NULL);
    
    if (broker)  {
-      mi=(f)(broker,ctx);
-      _SFCB_RETURN(mi);
+     mi=(f)(broker,ctx, status);
+     if (mi && status->rc == CMPI_RC_OK)
+       _SFCB_RETURN(mi);
    }
-   _SFCB_RETURN((CMPIQualifierDeclMI*)(void*)-1);
+   _SFCB_RETURN(NULL); 
 };
 
 /****************************************************************************/
@@ -329,7 +366,7 @@ int memAdd(void *ptr, int *memId);
 void *tool_mm_load_lib(const char *libname)
 {
    char filename[255];
-   sprintf(filename, "lib%s.so", libname);
+   snprintf(filename, 255, "lib%s.so", libname);
    return dlopen(filename, RTLD_LAZY);
 }
 
@@ -380,18 +417,18 @@ static void __cleanup_mt(void *ptr)
    _SFCB_EXIT();
 }
 
+static managed_thread *__memInit(int dontforce);
 
 /**
  * Initializes the current thread by adding it to the memory management sytem.
  */
 
-#include <pthread.h>
-
 void uninitGarbageCollector()
 {
-   managed_thread *mt=NULL;   
-   mt = (managed_thread *)
-       CMPI_BrokerExt_Ftab->getThreadSpecific(__mm_key);
+
+   /* do an init (don't force) just in case; otherwise, we risk using garbage */
+   managed_thread *mt =  __memInit(1);
+
    if (mt==NULL) return;
    
    __cleanup_mt(mt);
@@ -737,7 +774,11 @@ char *decode64(char *din)
    unsigned char *data=(unsigned char*)din;
    int i, o = 0, len = strlen((char*)data);
    unsigned char c, c1;
-   unsigned char *ret = (unsigned char *) malloc(len * 2);
+   unsigned char *ret = NULL;
+   if (len > 0)
+   {
+      ret = (unsigned char *) malloc(len * 2);
+   }
 
  for (i = 0; i < len; ++i) {
       c = (char) find(cvt, data[i]);
@@ -764,7 +805,10 @@ char *decode64(char *din)
       }
    }
 
-   ret[o] = 0;
+   if (ret)
+   {
+      ret[o] = 0;
+   }
    return ((char*)ret);
 }
 
@@ -882,7 +926,7 @@ void dumpTiming(int pid)
    
    if (collectStat==0) return;
    
-   sprintf(buffer,"/proc/%d/stat",pid);
+   snprintf(buffer, 4096, "/proc/%d/stat",pid);
    f=fopen(buffer,"r");
    l=fread(buffer,1,4095,f);
    fclose(f);   
@@ -919,3 +963,25 @@ double timevalDiff(struct timeval *sv, struct timeval *ev)
    return e-s;
 }
    
+char* sfcb_snprintf(const char* fmt, ...)
+{
+    va_list ap;
+    int len;
+    va_start(ap, fmt);
+    len = vsnprintf(NULL, 0, fmt, ap);
+    va_end(ap);
+    if (len <= 0)
+    {
+        return NULL;
+    }
+    char* str = (char*)malloc(len+1);
+    if (str == NULL)
+    {
+        return NULL;
+    }
+    va_start(ap, fmt);
+    vsnprintf(str, len+1, fmt, ap);
+    va_end(ap);
+    return str;
+}
+
